# type: ignore[misc]
"""yt-dlg module responsible for parsing the options. """
from __future__ import annotations

from pathlib import Path
from typing import Any

from .utils import remove_shortcuts


class OptionHolder:
    """Simple data structure that holds informations for the given option.

    Args:
        name (str): Option name. Must be a valid option name
            from the optionsmanager.OptionsManager class.
            See optionsmanager.OptionsManager load_default() method.

        flag (str): The option command line switch.
            See https://github.com/ytdl-org/youtube-dl/#options

        default_value (Any): The option default value. Must be the same type
            with the corresponding option from the optionsmanager.OptionsManager
            class.

        requirements (list): The requirements for the given option. This
            argument is a list of strings with the name of all the options
            that this specific option needs. For example 'subs_lang' needs the
            'write_subs' option to be enabled.

    """

    def __init__(
        self,
        name: str,
        flag: str,
        default_value: str | int | bool,
        requirements: list[str] | None = None,
    ):
        self.name = name
        self.flag = flag
        self.default_value = default_value
        self.requirements = requirements

    def is_boolean(self) -> bool:
        """Returns True if the option is a boolean switch else False."""
        return type(self.default_value) is bool

    def check_requirements(self, options_dict: dict[str, Any]) -> bool:
        """Check if the required options are enabled.

        Args:
            options_dict (dict): Dictionary with all the options.

        Returns:
            True if any of the required options is enabled else False.

        """
        if self.requirements is None:
            return True

        return any(options_dict[req] for req in self.requirements)


class OptionsParser:
    """Parse optionsmanager.OptionsManager options.

    This class is responsible for turning some of the youtube-dlg options
    to youtube-dl command line options.

    """

    def __init__(self):
        self._ydl_options = [
            OptionHolder("playlist_start", "--playlist-start", 1),
            OptionHolder("playlist_end", "--playlist-end", 0),
            OptionHolder("max_downloads", "--max-downloads", 0),
            OptionHolder("username", "-u", ""),
            OptionHolder("password", "-p", ""),
            OptionHolder("video_password", "--video-password", ""),
            OptionHolder("retries", "-R", 10),
            OptionHolder("proxy", "--proxy", ""),
            OptionHolder("user_agent", "--user-agent", ""),
            OptionHolder("referer", "--referer", ""),
            OptionHolder("ignore_errors", "-i", False),
            OptionHolder("write_description", "--write-description", False),
            OptionHolder("write_info", "--write-info-json", False),
            OptionHolder("write_thumbnail", "--write-thumbnail", False),
            OptionHolder("min_filesize", "--min-filesize", 0),
            OptionHolder("max_filesize", "--max-filesize", 0),
            OptionHolder("write_all_subs", "--all-subs", False),
            OptionHolder("write_auto_subs", "--write-auto-sub", False),
            OptionHolder("write_subs", "--write-sub", False),
            OptionHolder("keep_video", "-k", False),
            OptionHolder("restrict_filenames", "--restrict-filenames", False),
            OptionHolder("save_path", "-o", ""),
            OptionHolder(
                "embed_subs", "--embed-subs", False, ["write_auto_subs", "write_subs"]
            ),
            OptionHolder("to_audio", "-x", False),
            OptionHolder("audio_format", "--audio-format", ""),
            OptionHolder("video_format", "-f", "0"),
            OptionHolder("subs_lang", "--sub-lang", "", ["write_subs"]),
            OptionHolder("audio_quality", "--audio-quality", "5", ["to_audio"]),
            OptionHolder("youtube_dl_debug", "-v", False),
            OptionHolder("ignore_config", "--ignore-config", False),
            OptionHolder("native_hls", "--hls-prefer-native", False),
            OptionHolder("nomtime", "--no-mtime", False),
            OptionHolder("embed_thumbnail", "--embed-thumbnail", False),
            OptionHolder("add_metadata", "--add-metadata", False),
        ]

    def parse(self, options_dictionary: dict[str, Any]) -> list[str]:
        """Parse optionsmanager.OptionsManager options.

        Parses the given options to youtube-dl command line arguments.

        Args:
            options_dictionary (dict): Dictionary with all the options.

        Returns:
            List of strings with all the youtube-dl command line options.

        """
        options_list: list[str] = ["--newline"]

        # Create a copy of options_dictionary
        # We don't want to edit the original options dictionary
        # and change some of the options values like 'save_path' etc..
        options_dict = options_dictionary.copy()

        self._build_savepath(options_dict)
        self._build_videoformat(options_dict)
        self._build_filesizes(options_dict)

        # Parse basic youtube-dl command line options
        for option in self._ydl_options:
            # NOTE Special case should be removed
            if option.name == "to_audio" and not options_dict["audio_format"]:
                value = options_dict[option.name]

                if value != option.default_value:
                    options_list.append(option.flag)
            elif option.name == "audio_format":
                value = options_dict[option.name]

                if value != option.default_value:
                    if "-x" not in options_list:
                        options_list.append("-x")
                    options_list.append(option.flag)
                    options_list.append(str(value))

                    # NOTE Temp fix
                    # If current 'audio_quality' is not the default one ('5')
                    # then append the audio quality flag and value to the
                    # options list
                    if options_dict["audio_quality"] != "5":
                        options_list.append("--audio-quality")
                        options_list.append(str(options_dict["audio_quality"]))
            elif option.name == "audio_quality":
                # If the '--audio-quality' is not already in the options list
                # from the above branch then follow the standard procedure.
                # We don't have to worry for the sequence in which the code
                # will be executed since the 'audio_quality' option is placed
                # after the 'audio_format' option in the self._ydl_options list
                if option.flag not in options_list and option.check_requirements(
                    options_dict
                ):
                    value = options_dict[option.name]
                    assert 0 <= int(value) <= 9

                    options_list.append(option.flag)
                    options_list.append(str(value))
            elif option.check_requirements(options_dict):
                value = options_dict[option.name]

                if value != option.default_value:
                    options_list.append(option.flag)

                    if not option.is_boolean():
                        options_list.append(str(value))

        self.parse_cmd_args(options_dict, options_list)

        return options_list

    @staticmethod
    def parse_cmd_args(options_dict: dict[str, Any], options_list: list[str]) -> None:
        """
        Parse cmd_args

        Args:
            options_dict (dict): Reference to options dictionary.
            options_list (list): Reference to options list to parse.

        """
        # TODO: Handle special cases of single and doble queotes in options better

        # Indicates whether an item needs special handling
        special_case = False

        # Temp list to hold special items
        special_items = []

        for item in options_dict["cmd_args"].split():

            # Its a special case if its already a special case
            # or an item starts with single/double quotes
            special_case = special_case or item[0] in ['"', "'"]

            if special_case:
                special_items.append(item)
            else:
                options_list.append(item)

            # If its a special case and we meet a double quote
            # at the end of the item, special case is over and
            # we need to join, filter and append our special items
            # to the options list
            if special_case and item[-1] in ['"', "'"]:
                options_list.append(" ".join(special_items)[1:-1])

                special_case = False
                special_items = []

    @staticmethod
    def _build_savepath(options_dict: dict[str, str]) -> None:
        """Build the save path.

        We use this method to build the value of the 'save_path' option and
        store it back to the options dictionary.

        Args:
            options_dict (dict): Copy of the original options dictionary.

        """
        # Check OUTPUT_FORMATS values (str)
        assert isinstance(options_dict["output_format"], str)

        save_path: str = remove_shortcuts(options_dict["save_path"])

        if options_dict["output_format"] == "0":
            template: str = "%(id)s.%(ext)s"
        elif options_dict["output_format"] == "1":
            template = "%(title)s.%(ext)s"
        elif options_dict["output_format"] == "2":
            template = "%(title)s-%(id)s.%(ext)s"
        elif options_dict["output_format"] == "4":
            template = "%(title)s-%(height)sp.%(ext)s"
        elif options_dict["output_format"] == "5":
            template = "%(title)s-%(id)s-%(height)sp.%(ext)s"
        else:
            template = options_dict["output_template"]

        options_dict["save_path"] = str(Path(save_path) / Path(template))

    @staticmethod
    def _build_videoformat(options_dict: dict[str, str]) -> None:
        """Build the video format.

        We use this method to build the value of the 'video_format' option and
        store it back to the options dictionary.

        Args:
            options_dict (dict): Copy of the original options dictionary.

        """
        if (
            options_dict["video_format"] != "0"
            and options_dict["second_video_format"] != "0"
        ):
            options_dict[
                "video_format"
            ] = f'{options_dict["video_format"]}+{options_dict["second_video_format"]}'

    @staticmethod
    def _build_filesizes(options_dict: dict[str, str | int]) -> None:
        """Build the filesize options values.

        We use this method to build the values of 'min_filesize' and
        'max_filesize' options and store them back to options dictionary.

        Args:
            options_dict (dict): Copy of the original options dictionary.

        """
        min_filesize = options_dict.get("min_filesize")
        max_filesize = options_dict.get("max_filesize")

        if min_filesize:
            options_dict[
                "min_filesize"
            ] = f'{options_dict["min_filesize"]}{options_dict["min_filesize_unit"]}'

        if max_filesize:
            options_dict[
                "max_filesize"
            ] = f'{options_dict["max_filesize"]}{options_dict["max_filesize_unit"]}'



from .info import (
    __appname__,
    __author__,
    __contact__,
    __description__,
    __descriptionfull__,
    __license__,
    __licensefull__,
    __maintainer__,
    __maintainer_contact__,
    __packagename__,
    __projecturl__,
)
from .version import __version__



"""yt-dlg module responsible for handling the log stuff. """


import logging
from logging.handlers import RotatingFileHandler
from pathlib import Path

from .utils import check_path, get_encoding  # type: ignore[attr-defined]


class LogManager:
    """Simple log manager for youtube-dl.

    This class is mainly used to log the youtube-dl STDERR.

    Attributes:
        LOG_FILENAME (str): Filename of the log file.
        MAX_LOGSIZE (int): Maximum size(Bytes) of the log file.

    Args:
        config_path (str): Absolute path where LogManager should
            store the log file.

        add_time (bool): If True LogManager will also log the time.

    """

    LOG_FILENAME = "log"
    MAX_LOGSIZE = 524288  # Bytes

    def __init__(self, config_path: str, add_time: bool = False):
        self.config_path: str = config_path
        self.add_time: bool = add_time
        self.log_file: str = str(Path(config_path) / Path(self.LOG_FILENAME))
        self._encoding: str = get_encoding()
        self.logger: logging.Logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.DEBUG)

        check_path(self.config_path)

        self.handler = RotatingFileHandler(
            filename=self.log_file,
            maxBytes=LogManager.MAX_LOGSIZE,
            backupCount=5,
            encoding=self._encoding,
        )

        fmt = "%(levelname)s-%(threadName)s-%(message)s"

        if self.add_time:
            fmt = "%(asctime)s-" + fmt

        self.handler.setFormatter(logging.Formatter(fmt=fmt))
        self.logger.addHandler(self.handler)

    def log_size(self) -> int:
        """Return log file size in Bytes."""
        path = Path(self.log_file)

        if not path.exists():
            return 0

        return path.stat().st_size

    def clear(self) -> None:
        """Clear log file."""
        with open(self.log_file, "w") as log:
            log.write("")

    def log(self, data: str) -> None:
        """Log data to the log file.

        Args:
            data (str): String to write to the log file.

        """
        self.logger.debug(str(data))



"""yt-dlg setup file.

Examples:
    Windows/Linux::

        python setup.py pyinstaller

    Linux::

        python setup.py install

    Build source distribution::

        python setup.py sdist

    Build platform distribution::

        python setup.py bdist

    Build the translations::

        python setup.py build_trans

    Build with updates disabled::

        python setup.py no_updates


    Basic steps of the setup::

        * Run pre-build tasks
        * Call setup handler based on OS & options
            * Set up hicolor icons (if supported by platform)
            * Set up fallback pixmaps icon (if supported by platform)
            * Set up package level pixmaps icons (*.png)
            * Set up package level i18n files (*.mo)
            * Set up scripts (executables) (if supported by platform)
        * Run setup

"""

import glob
import os
import sys
import time
from distutils import log
from distutils.spawn import spawn

from setuptools import Command, setup

__packagename__ = "youtube_dl_gui"
__packageytdlg__ = "yt_dlg"

PYINSTALLER = len(sys.argv) >= 2 and sys.argv[1] == "pyinstaller"

try:
    from PyInstaller import compat as pyi_compat

    if pyi_compat.is_win:
        # noinspection PyUnresolvedReferences
        from PyInstaller.utils.win32.versioninfo import (
            FixedFileInfo,
            SetVersion,
            StringFileInfo,
            StringStruct,
            StringTable,
            VarFileInfo,
            VarStruct,
            VSVersionInfo,
        )
except ImportError:
    pyi_compat = None
    if PYINSTALLER:
        print("Cannot import pyinstaller", file=sys.stderr)
        exit(1)

# Get the version from youtube_dl_gui/version.py without importing the package
exec(
    compile(
        open(__packagename__ + "/version.py").read(),
        __packagename__ + "/version.py",
        "exec",
    )
)
# Get the info from youtube_dl_gui/info.py without importing the package
exec(
    compile(
        open(__packagename__ + "/info.py").read(), __packagename__ + "/info.py", "exec"
    )
)

DESCRIPTION = __description__
LONG_DESCRIPTION = open("README.md", encoding="utf-8").read()


def on_windows():
    """Returns True if OS is Windows."""
    return os.name == "nt"


def version2tuple(commit=0):
    version_list = str(__version__).split(".")

    if len(version_list) > 3:
        del version_list[3]

    _release = commit

    _year, _month, _day = tuple(map(int, version_list))
    return _year, _month, _day, _release


def version2str(commit=0):
    version_tuple = version2tuple(commit)
    return "%s.%s.%s.%s" % version_tuple


# noinspection PyAttributeOutsideInit,PyArgumentList
class BuildTranslations(Command):
    description = "Build the translation files"
    user_options = []

    def initialize_options(self):
        self.search_pattern = None

    def finalize_options(self):
        self.search_pattern = os.path.join(
            __packagename__, "locale", "*", "LC_MESSAGES", "youtube_dl_gui.po"
        )

    def run(self):
        import polib

        po_file = ""

        try:
            for po_file in glob.glob(self.search_pattern):
                mo_file = po_file.replace(".po", ".mo")
                po = polib.pofile(po_file)

                log.info(f"Building MO file for '{po_file}'")
                po.save_as_mofile(mo_file)
        except OSError as error:
            log.error(f"{error}, exiting...")
            sys.exit(1)


# noinspection PyAttributeOutsideInit,PyArgumentList
class BuildNoUpdate(Command):
    description = "Build with updates disabled"
    user_options = []

    def initialize_options(self):
        self.build_lib = os.path.dirname(os.path.abspath(__file__))

    def finalize_options(self):
        pass

    def run(self):
        self.__disable_updates()

    def __disable_updates(self):
        lib_dir = os.path.join(self.build_lib, __packagename__)
        target_file = "optionsmanager.py"
        # Options file should be available from previous build commands
        optionsfile = os.path.join(lib_dir, target_file)

        with open(optionsfile) as input_file:
            data = input_file.readlines()

        if data is None:
            log.error("Building with updates disabled failed!")
            sys.exit(1)

        for index, line in enumerate(data):
            if '"disable_update": False' in line:
                log.info("Disabling updates...")
                data[index] = line.replace("False", "True")
                break

        with open(optionsfile, "w") as output_file:
            output_file.writelines(data)


class BuildPyinstallerBin(Command):
    description = "Build the executable"
    user_options = []
    version_file = None
    if pyi_compat and pyi_compat.is_win:
        version_file = VSVersionInfo(
            ffi=FixedFileInfo(
                filevers=version2tuple(),
                prodvers=version2tuple(),
                mask=0x3F,
                flags=0x0,
                OS=0x4,
                fileType=0x1,
                subtype=0x0,
                date=(0, 0),
            ),
            kids=[
                VarFileInfo([VarStruct("Translation", [0, 1200])]),
                StringFileInfo(
                    [
                        StringTable(
                            "000004b0",
                            [
                                StringStruct("CompanyName", __maintainer_contact__),
                                StringStruct("FileDescription", DESCRIPTION),
                                StringStruct("FileVersion", version2str()),
                                StringStruct("InternalName", "yt-dlg.exe"),
                                StringStruct(
                                    "LegalCopyright",
                                    __projecturl__ + "LICENSE",
                                ),
                                StringStruct("OriginalFilename", "yt-dlg.exe"),
                                StringStruct("ProductName", __appname__),
                                StringStruct("ProductVersion", version2str()),
                            ],
                        )
                    ]
                ),
            ],
        )

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self, version=version_file):
        """Run pyinstaller"""
        path_sep = ";" if on_windows() else ":"
        spawn(
            [
                "pyinstaller",
                "-w",
                "-F",
                "--icon=" + __packagename__ + "/data/pixmaps/youtube-dl-gui.ico",
                "--add-data="
                + __packagename__
                + "/data"
                + path_sep
                + __packagename__
                + "/data",
                "--add-data="
                + __packagename__
                + "/locale"
                + path_sep
                + __packagename__
                + "/locale",
                "--exclude-module=tests",
                "--name=yt-dlg",
                __packagename__ + "/__main__.py",
            ],
            dry_run=self.dry_run,
        )

        if version:
            time.sleep(3)
            SetVersion("./dist/yt-dlg.exe", version)


cmdclass = {
    "build_trans": BuildTranslations,
    "no_updates": BuildNoUpdate,
}

# Add pixmaps icons (*.png) & i18n files
package_data = {__packagename__: ["data/pixmaps/*.png", "locale/*/LC_MESSAGES/*.mo"]}


def setup_linux():
    """Setup params for Linux"""
    data_files_linux = []
    # Add hicolor icons
    for path in glob.glob("youtube_dl_gui/data/icons/hicolor/*x*"):
        size = os.path.basename(path)

        dst = f"share/icons/hicolor/{size}/apps"
        src = f"{path}/apps/youtube-dl-gui.png"

        data_files_linux.append((dst, [src]))
    # Add fallback icon, see issue #14
    data_files_linux.append(
        ("share/pixmaps", ["youtube_dl_gui/data/pixmaps/youtube-dl-gui.png"])
    )
    # Add man page
    data_files_linux.append(("share/man/man1", ["youtube-dl-gui.1"]))
    return {
        "data_files": data_files_linux,
        "package_data": package_data,
    }


def setup_windows():
    """Setup params for Windows"""
    return {
        "package_data": package_data,
    }


params = {}

if PYINSTALLER:
    cmdclass.update({"pyinstaller": BuildPyinstallerBin})
else:
    params = setup_windows() if on_windows() else setup_linux()
    params["entry_points"] = {
        "console_scripts": ["yt-dlg = " + __packagename__ + ".app:main"]
    }


setup(
    name=__packageytdlg__,
    version=__version__,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url=__projecturl__,
    download_url=__githuburl__ + "releases/latest",
    project_urls={
        # "Documentation": "https://yt-dlg.readthedocs.io",
        "Source": __githuburl__,
        "Tracker": __githuburl__ + "issues",
        "Funding": __projecturl__ + "donate.html",
    },
    author=__author__,
    author_email=__contact__,
    maintainer=__maintainer__,
    maintainer_email=__maintainer_contact__,
    license=__license__,
    packages=[__packagename__],
    install_requires=[
        "pypubsub>=4.0.3",
        "polib>=1.1.0",
        "wxPython<=4.1.2a1,>=4.0.7.post2",
        "pyinstaller<=4.6,>=3.6",
    ],
    python_requires=">=3.7",
    classifiers=[
        "Topic :: Multimedia :: Video",
        "Development Status :: 5 - Production/Stable",
        "Environment :: MacOS X :: Cocoa",
        "Environment :: Win32 (MS Windows)",
        "Environment :: X11 Applications :: GTK",
        "License :: Public Domain",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    cmdclass=cmdclass,
    **params,
)



# type: ignore[misc]
"""yt-dlg module responsible for the main app window. """
from __future__ import annotations

import os
from pathlib import Path
from typing import Any, Callable, Iterable

import wx
import wx.adv

# noinspection PyPep8Naming
from pubsub import pub as Publisher
from wx.lib.mixins.listctrl import ListCtrlAutoWidthMixin

from .darktheme import dark_mode
from .downloadmanager import (
    MANAGER_PUB_TOPIC,
    WORKER_PUB_TOPIC,
    DownloadItem,
    DownloadList,
    DownloadManager,
)
from .formats import AUDIO_FORMATS, DEFAULT_FORMATS, FORMATS, VIDEO_FORMATS
from .info import (
    __appname__,
    __author__,
    __descriptionfull__,
    __licensefull__,
    __maintainer__,
    __projecturl__,
)
from .logmanager import LogManager
from .optionsframe import OptionsFrame
from .optionsmanager import OptionsManager
from .parsers import OptionsParser
from .updatemanager import UPDATE_PUB_TOPIC, UpdateThread
from .utils import (
    YOUTUBEDL_BIN,
    build_command,
    get_icon_file,
    get_key,
    get_pixmaps_dir,
    get_time,
    open_file,
    shutdown_sys,
)
from .version import __version__
from .widgets import (
    ButtonsChoiceDialog,
    ClipDialog,
    ExtComboBox,
    ListBoxComboPopup,
    LogGUI,
    MessageDialog,
    ShutdownDialog,
)

_: Callable[[str], str] = wx.GetTranslation


class ListCtrl(wx.ListCtrl, ListCtrlAutoWidthMixin):

    """Custom ListCtrl widget.

    Args:
        columns (dict): See MainFrame class STATUSLIST_COLUMNS attribute.

    """

    def __init__(self, columns: dict[str, tuple[int, str, int, bool]], *args, **kwargs):
        super().__init__(*args, **kwargs)
        ListCtrlAutoWidthMixin.__init__(self)
        self.columns = columns
        self._list_index = 0
        self._map_id: dict[int, int] = {}
        self._url_list: set[str] = set()
        self._set_columns()

    def remove_row(self, row_number: int):
        self.DeleteItem(row_number)
        total = len(self._map_id)
        for row in range(row_number, total - 1):
            self._map_id[row] = self._map_id[row + 1]
        del self._map_id[total - 1]
        self._list_index -= 1

    def move_item_up(self, row_number: int):
        self._move_item(row_number, row_number - 1)

    def move_item_down(self, row_number: int):
        self._move_item(row_number, row_number + 1)

    def _move_item(self, cur_row: int, new_row: int):
        self.Freeze()
        item = self.GetItem(cur_row)
        self.DeleteItem(cur_row)

        item.SetId(new_row)
        self.InsertItem(item)
        # Swap Data associated (Python Data Mixing)
        self._map_id[new_row], self._map_id[cur_row] = (
            self._map_id[cur_row],
            self._map_id[new_row],
        )

        self.Select(new_row)
        self.Thaw()
        # self.SetFocus()

    def has_url(self, url: str):
        """Returns True if the url is aleady in the ListCtrl else False.

        Args:
            url (string): URL string.

        """
        return url in self._url_list

    def bind_item(self, download_item: DownloadItem):
        self.InsertItem(self._list_index, download_item.url)

        self.SetItemData(self._list_index, download_item.object_id)
        self._map_id[self._list_index] = download_item.object_id

        self._update_from_item(self._list_index, download_item)

        self._list_index += 1

    def GetItemData(self, row_index_selected: int) -> int | None:
        return self._map_id.get(row_index_selected, None)

    def _update_from_item(self, row: int, download_item: DownloadItem):
        progress_stats = download_item.progress_stats

        for key in self.columns:
            column = self.columns[key][0]

            if key == "status" and progress_stats["playlist_index"]:
                # Not the best place but we build the playlist status here
                status = (
                    f'{progress_stats["status"]} '
                    f'{progress_stats["playlist_index"]}/{progress_stats["playlist_size"]}'
                )

                self.SetItem(row, column, status)
            else:
                self.SetItem(row, column, progress_stats[key])

    def clear(self):
        """Clear the ListCtrl widget & reset self._list_index and
        self._url_list."""
        self.DeleteAllItems()
        self._list_index = 0
        self._url_list = set()

    def is_empty(self):
        """Returns True if the list is empty else False."""
        return self._list_index == 0

    def get_selected(self) -> int:
        return self.GetNextItem(-1, wx.LIST_NEXT_ALL, wx.LIST_STATE_SELECTED)

    def get_all_selected(self) -> list[int]:
        return [index for index in range(self._list_index) if self.IsSelected(index)]

    def deselect_all(self):
        for index in range(self._list_index):
            self.Select(index, on=0)

    def get_next_selected(self, start: int = -1, reverse: bool = False) -> int:
        if start == -1:
            start = self._list_index - 1 if reverse else 0
        elif reverse:
            start -= 1
        else:
            start += 1

        end = -1 if reverse else self._list_index
        step = -1 if reverse else 1

        for index in range(start, end, step):
            if self.IsSelected(index):
                return index

        return -1

    def _set_columns(self):
        """Initializes ListCtrl columns.
        See MainFrame STATUSLIST_COLUMNS attribute for more info."""
        for column_item in sorted(self.columns.values()):
            self.InsertColumn(column_item[0], column_item[1], width=wx.LIST_AUTOSIZE)

            # If the column width obtained from wxLIST_AUTOSIZE
            # is smaller than the minimum allowed column width
            # then set the column width to the minimum allowed size
            if self.GetColumnWidth(column_item[0]) < column_item[2]:
                self.SetColumnWidth(column_item[0], column_item[2])

            # Set auto-resize if enabled
            if column_item[3]:
                self.setResizeColumn(column_item[0])


class MainFrame(wx.Frame):
    """Main window class.

    This class is responsible for creating the main app window
    and binding the events.

    Attributes:
        FRAMES_MIN_SIZE (tuple): Tuple that contains the minumum width, height of the frame.

    Args:
        opt_manager (optionsmanager.OptionsManager): Object responsible for
            handling the settings.

        log_manager (logmanager.LogManager): Object responsible for handling
            the log stuff.

        _parent (wx.Window): Frame parent.

    """

    FRAMES_MIN_SIZE = (560, 360)

    def __init__(
        self,
        opt_manager: OptionsManager,
        log_manager: LogManager | None = None,
        _parent: wx.Window | None = None,
    ):
        super().__init__(
            _parent,
            wx.ID_ANY,
            __appname__,
            size=opt_manager.options.get("main_win_size", OptionsManager.MAIN_WIN_SIZE),
        )

        # Labels area: Strings for the widgets labels.
        self.URLS_LABEL = _("Enter URLs below")
        self.UPDATE_LABEL = _("Update")
        self.SETTINGS_LABEL = _("Setting")
        self.OPTIONS_LABEL = _("Options")
        self.STOP_LABEL = _("Stop")
        self.INFO_LABEL = _("Info")
        self.WELCOME_MSG = _("Welcome")
        self.WARNING_LABEL = _("Warning")
        self.ADD_LABEL = _("Add")
        self.DOWNLOAD_LIST_LABEL = _("Download list")
        self.DELETE_LABEL = _("Delete")
        self.PLAY_LABEL = _("Play")
        self.UP_LABEL = _("Up")
        self.DOWN_LABEL = _("Down")
        self.RELOAD_LABEL = _("Reload")
        self.PAUSE_LABEL = _("Pause")
        self.START_LABEL = _("Start")
        self.ABOUT_LABEL = _("About")
        self.VIEWLOG_LABEL = _("View Log")

        self.SUCC_REPORT_MSG = _(
            "Downloaded {0} URL(s) in {1} "
            "day(s) {2} hour(s) {3} minute(s) {4} second(s)"
        )
        self.DL_COMPLETED_MSG = _("Downloads completed")
        self.URL_REPORT_MSG = _(
            "Total Progress: {0:.1f}% | Queued ({1}) Paused ({2}) Active ({3}) "
            "Completed ({4}) Error ({5})"
        )
        self.CLOSING_MSG = _("Stopping downloads")
        self.CLOSED_MSG = _("Downloads stopped")
        self.PROVIDE_URL_MSG = _("You need to provide at least one URL")
        self.DOWNLOAD_STARTED = _("Downloads started")
        self.CHOOSE_DIRECTORY = _("Choose Directory")

        self.DOWNLOAD_ACTIVE = _(
            "Download in progress. Please wait for all downloads to complete"
        )
        self.UPDATE_ACTIVE = _("Update already in progress")

        self.UPDATING_MSG = _("Downloading latest CLI Backend. Please wait...")
        self.UPDATE_ERR_MSG = _("CLI Backend download failed [{0}]")
        self.UPDATE_SUCC_MSG = _("Successfully downloaded CLI Backend.")

        self.OPEN_DIR_ERR = _(
            "Unable to open directory: '{dir}'. " "The specified path does not exist"
        )
        self.SHUTDOWN_ERR = _(
            "Error while shutting down. " "Make sure you typed the correct password"
        )
        self.SHUTDOWN_MSG = _("Shutting down system")

        self.VIDEO_LABEL = _("Title")
        self.EXTENSION_LABEL = _("Extension")
        self.SIZE_LABEL = _("Size")
        self.PERCENT_LABEL = _("Percent")
        self.ETA_LABEL = _("ETA")
        self.SPEED_LABEL = _("Speed")
        self.STATUS_LABEL = _("Status")

        # STATUSLIST_COLUMNS
        #
        # Dictionary which contains the columns for the wxListCtrl widget.
        # Each key represents a column and holds informations about itself.
        # Structure informations:
        #  column_key: (column_number, column_label, minimum_width, is_resizable)
        #
        self.STATUSLIST_COLUMNS: dict[str, tuple[int, str, int, bool]] = {
            "filename": (0, self.VIDEO_LABEL, 180, True),
            "extension": (1, self.EXTENSION_LABEL, 80, False),
            "filesize": (2, self.SIZE_LABEL, 80, False),
            "percent": (3, self.PERCENT_LABEL, 80, False),
            "eta": (4, self.ETA_LABEL, 90, False),
            "speed": (5, self.SPEED_LABEL, 90, False),
            "status": (6, self.STATUS_LABEL, 90, False),
        }

        self.opt_manager = opt_manager
        self.log_manager = log_manager
        self.download_manager: DownloadManager | None = None
        self.update_thread: UpdateThread | None = None
        self.app_icon: wx.Icon | None = None

        self._download_list = DownloadList()

        # Set up youtube-dl options parser
        self._options_parser = OptionsParser()

        # Get the pixmaps directory
        self._pixmaps_path: str | None = get_pixmaps_dir()

        # Set the Timer
        self._app_timer = wx.Timer(self)

        # Set the app icon
        app_icon_path: str | None = get_icon_file()
        if app_icon_path:
            self.app_icon = wx.Icon(app_icon_path, wx.BITMAP_TYPE_PNG)
            self.SetIcon(self.app_icon)

        bitmap_data = (
            ("down", "arrow_down_32px.png"),
            ("up", "arrow_up_32px.png"),
            ("play", "camera_32px.png"),
            ("start", "cloud_download_32px.png"),
            ("delete", "delete_32px.png"),
            ("folder", "folder_32px.png"),
            ("pause", "pause_32px.png"),
            ("resume", "play_arrow_32px.png"),
            ("reload", "reload_32px.png"),
            ("settings", "settings_20px.png"),
            ("stop", "stop_32px.png"),
        )

        self._bitmaps: dict[str, wx.Bitmap] = {
            target: wx.Bitmap(str(Path(self._pixmaps_path or ".").joinpath(name)))
            for target, name in bitmap_data
        }

        # Set the data for all the wx.Button items
        # name, label, size, event_handler
        buttons_data = (
            ("delete", self.DELETE_LABEL, (-1, -1), self._on_delete, wx.BitmapButton),
            ("play", self.PLAY_LABEL, (-1, -1), self._on_play, wx.BitmapButton),
            ("up", self.UP_LABEL, (-1, -1), self._on_arrow_up, wx.BitmapButton),
            ("down", self.DOWN_LABEL, (-1, -1), self._on_arrow_down, wx.BitmapButton),
            ("reload", self.RELOAD_LABEL, (-1, -1), self._on_reload, wx.BitmapButton),
            ("pause", self.PAUSE_LABEL, (-1, -1), self._on_pause, wx.BitmapButton),
            ("start", self.START_LABEL, (-1, -1), self._on_start, wx.BitmapButton),
            (
                "settings",
                self.SETTINGS_LABEL,
                (-1, -1),
                self._on_settings,
                wx.BitmapButton,
            ),
            ("savepath", "...", (35, -1), self._on_savepath, wx.Button),
            ("add", self.ADD_LABEL, (-1, -1), self._on_add, wx.Button),
        )

        # Set the data for the settings menu item
        # label, event_handler
        settings_menu_data = (
            (self.OPTIONS_LABEL, self._on_options),
            (self.UPDATE_LABEL, self._on_update),
            (self.VIEWLOG_LABEL, self._on_viewlog),
            (self.ABOUT_LABEL, self._on_about),
        )

        statuslist_menu_data = (
            (_("Get URL"), self._on_geturl),
            (_("Get command"), self._on_getcmd),
            (_("Clip Multimedia"), self._on_clip),
            (_("Open destination"), self._on_open_dest),
            (_("Re-enter"), self._on_reenter),
        )

        # Is Dark Theme
        self._dark_mode = self.opt_manager.options.get("dark_mode", False)

        # Create options frame
        self._options_frame = OptionsFrame(self, self._dark_mode)

        # Create frame components
        self._panel = wx.Panel(self)

        self._url_text = self._create_statictext(self.URLS_LABEL)

        self._url_list = self._create_textctrl(
            wx.TE_MULTILINE | wx.TE_DONTWRAP, self._on_urllist_edit
        )

        self._folder_icon = self._create_static_bitmap(
            self._bitmaps["folder"], self._on_open_path
        )

        self._path_combobox = ExtComboBox(self._panel, 5, style=wx.CB_READONLY)

        FORMATS["0"] = _("default")
        DEFAULT_FORMATS["0"] = _("default")

        self._videoformat_combobox = wx.ComboCtrl(
            self._panel, size=(180, -1), style=wx.CB_READONLY
        )
        self._popup_ctrl = ListBoxComboPopup(
            self._videoformat_combobox, self._dark_mode
        )
        self._videoformat_combobox.SetPopupControl(self._popup_ctrl)

        self._download_text = self._create_statictext(self.DOWNLOAD_LIST_LABEL)
        self._status_list = ListCtrl(
            self.STATUSLIST_COLUMNS,
            parent=self._panel,
            style=wx.LC_REPORT | wx.LC_HRULES | wx.LC_VRULES,
        )

        # Dictionary to store all the buttons
        self._buttons = {}

        for item in buttons_data:
            name, label, size, evt_handler, parent = item

            button = parent(self._panel, size=size)

            if parent == wx.Button:
                button.SetLabel(label)
            elif parent == wx.BitmapButton:
                button.SetToolTip(wx.ToolTip(label))

            if name in self._bitmaps:
                button.SetBitmap(self._bitmaps[name], wx.TOP)

            if evt_handler is not None:
                button.Bind(wx.EVT_BUTTON, evt_handler)

            self._buttons[name] = button

        self._status_bar = self.CreateStatusBar()

        # Create extra components
        self._settings_menu = self._create_menu_item(settings_menu_data)
        self._statuslist_menu = self._create_menu_item(statuslist_menu_data)

        # Overwrite the menu hover event to avoid changing the statusbar
        self.Bind(wx.EVT_MENU_HIGHLIGHT, lambda event: None)

        # Bind extra events
        self.Bind(
            wx.EVT_LIST_ITEM_RIGHT_CLICK,
            self._on_statuslist_right_click,
            self._status_list,
        )
        self.Bind(wx.EVT_TEXT, self._update_savepath, self._path_combobox)
        self.Bind(
            wx.EVT_LIST_ITEM_SELECTED, self._update_pause_button, self._status_list
        )
        self.Bind(
            wx.EVT_LIST_ITEM_DESELECTED, self._update_pause_button, self._status_list
        )
        self.Bind(wx.EVT_CLOSE, self._on_close)
        self.Bind(wx.EVT_TIMER, self._on_timer, self._app_timer)

        self._videoformat_combobox.Bind(wx.EVT_COMBOBOX, self._update_videoformat)
        self._videoformat_combobox.Bind(
            wx.EVT_COMBOBOX_CLOSEUP, self._update_videoformat
        )

        # Set threads wxCallAfter handlers
        self._set_publisher(self._update_handler, UPDATE_PUB_TOPIC)
        self._set_publisher(self._download_worker_handler, WORKER_PUB_TOPIC)
        self._set_publisher(self._download_manager_handler, MANAGER_PUB_TOPIC)

        # Set up extra stuff
        self.Center()
        self.SetMinSize(self.FRAMES_MIN_SIZE)

        self._status_bar_write(self.WELCOME_MSG)

        self._update_videoformat_combobox()
        self._path_combobox.LoadMultiple(
            self.opt_manager.options.get("save_path_dirs", [])
        )
        self._path_combobox.SetValue(self.opt_manager.options.get("save_path", "."))
        self._update_savepath(None)

        self._set_layout()
        # Set Dark Theme
        dark_mode(self._panel, self._dark_mode)
        dark_mode(self._options_frame.panel, self._dark_mode)

        if os.name == "nt":
            self._videoformat_combobox.SetForegroundColour("Black")

        self._url_list.SetFocus()

    @staticmethod
    def _create_menu_item(items: Iterable[tuple[str, Any]]) -> wx.Menu:
        menu = wx.Menu()

        for label, evt_handler in items:
            menu_item = menu.Append(-1, label)

            menu.Bind(wx.EVT_MENU, evt_handler, menu_item)

        return menu

    def _on_statuslist_right_click(self, event):
        selected = event.GetIndex()

        if selected != -1:
            self._status_list.deselect_all()
            self._status_list.Select(selected, on=1)

            self.PopupMenu(self._statuslist_menu)

    # noinspection PyUnusedLocal
    def _on_reenter(self, event):
        selected = self._status_list.get_selected()

        if selected != -1:
            object_id: int | None = self._status_list.GetItemData(selected)
            download_item = self._download_list.get_item(object_id)

            if download_item and download_item.stage != "Active":
                self._status_list.remove_row(selected)
                self._download_list.remove(object_id)

                options = self._options_parser.parse(self.opt_manager.options)

                download_item = DownloadItem(download_item.url, options)
                download_item.path = self.opt_manager.options["save_path"]

                if not self._download_list.has_item(download_item.object_id):
                    self._status_list.bind_item(download_item)
                    self._download_list.insert(download_item)

    def reset(self):
        self._update_videoformat_combobox()
        self._path_combobox.LoadMultiple(self.opt_manager.options["save_path_dirs"])
        self._path_combobox.SetValue(self.opt_manager.options["save_path"])

    # noinspection PyUnusedLocal
    def _on_open_dest(self, event):
        selected = self._status_list.get_selected()

        if selected != -1:
            object_id = self._status_list.GetItemData(selected)
            download_item = self._download_list.get_item(object_id)

            if download_item.path:
                open_file(download_item.path)

    # noinspection PyUnusedLocal
    def _on_open_path(self, event):
        open_file(self._path_combobox.GetValue())

    def _copy_to_clipboard(self, text: str):
        clipdata = wx.TextDataObject()
        clipdata.SetText(text)
        wx.TheClipboard.Open()
        wx.TheClipboard.SetData(clipdata)
        wx.TheClipboard.Close()

    # noinspection PyUnusedLocal
    def _on_geturl(self, event):
        selected = self._status_list.get_selected()

        if selected != -1:
            object_id = self._status_list.GetItemData(selected)
            download_item = self._download_list.get_item(object_id)

            url = download_item.url

            if not wx.TheClipboard.IsOpened():
                self._copy_to_clipboard(url)

    # noinspection PyUnusedLocal
    def _on_getcmd(self, event):
        selected = self._status_list.get_selected()

        if selected != -1:
            object_id = self._status_list.GetItemData(selected)
            download_item = self._download_list.get_item(object_id)

            cmd = build_command(
                download_item.options,
                download_item.url,
                self.opt_manager.options.get("cli_backend", YOUTUBEDL_BIN),
            )

            if not wx.TheClipboard.IsOpened():
                self._copy_to_clipboard(cmd)

    def _on_clip(self, event):
        """
        Add external downloader args options
        to the end of the list options if ID_OK
        """
        index = self._status_list.get_next_selected()

        if index != -1:
            object_id = self._status_list.GetItemData(index)
            selected_download_item = self._download_list.get_item(object_id)

            if selected_download_item.stage == "Active":
                self._create_popup(
                    _("Item is active, can not set the multimedia clip"),
                    self.WARNING_LABEL,
                    wx.OK | wx.ICON_EXCLAMATION,
                )
                return

            dlg = ClipDialog(self, selected_download_item, self._dark_mode)
            check_options = dlg.CHECK_OPTIONS

            result = dlg.ShowModal() == wx.ID_OK

            clip_start: wx.TimeSpan = dlg.clip_start.GetValue(as_wxTimeSpan=True)
            clip_end: wx.TimeSpan = dlg.clip_end.GetValue(as_wxTimeSpan=True)

            dlg.Destroy()

            options: list[str] = selected_download_item.options

            if result:
                options.append(f"{check_options[0]}")
                options.append("ffmpeg")
                options.append(f"{check_options[1]}")
                options.append(
                    f"-ss {clip_start.GetSeconds()} -to {clip_end.GetSeconds()}"
                )

    # noinspection PyUnusedLocal
    def _on_timer(self, event):
        total_percentage = 0.0
        queued = paused = active = completed = error = 0

        for item in self._download_list.get_items():
            if item.stage == "Paused":
                paused += 1
            elif item.stage == "Queued":
                queued += 1
            if item.stage == "Active":
                active += 1
                total_percentage += float(item.progress_stats["percent"].split("%")[0])
            if item.stage == "Completed":
                completed += 1
            elif item.stage == "Error":
                error += 1

        # REFACTOR Store percentage as float in the DownloadItem?
        # REFACTOR DownloadList keep track for each item stage?

        items_count = active + completed + error + queued
        total_percentage += completed * 100.0 + error * 100.0

        if items_count:
            total_percentage /= items_count

        msg = self.URL_REPORT_MSG.format(
            total_percentage, queued, paused, active, completed, error
        )

        if self.update_thread is None:
            # Dont overwrite the update messages
            self._status_bar_write(msg)

    # noinspection PyUnusedLocal
    def _update_pause_button(self, event):
        selected_rows = self._status_list.get_all_selected()

        label = _("Pause")
        bitmap = self._bitmaps["pause"]

        for row in selected_rows:
            object_id = self._status_list.GetItemData(row)
            download_item = self._download_list.get_item(object_id)

            if download_item.stage == "Paused":
                # If we find one or more items in Paused
                # state set the button functionality to resume
                label = _("Resume")
                bitmap = self._bitmaps["resume"]
                break

        self._buttons["pause"].SetLabel(label)
        self._buttons["pause"].SetToolTip(wx.ToolTip(label))
        self._buttons["pause"].SetBitmap(bitmap, wx.TOP)

    def _get_listbox_headers(self):
        lb_popup_ctr: ListBoxComboPopup = self._popup_ctrl
        return lb_popup_ctr.GetControl()

    def _update_videoformat_combobox(self):
        lb_headers = self._get_listbox_headers()
        lb_headers.Clear()

        lb_headers.add_items(list(DEFAULT_FORMATS.values()), False)

        vformats: list[str] = [
            FORMATS[get_key(vformat, FORMATS)]
            for vformat in self.opt_manager.options.get("selected_video_formats", [])
        ]

        aformats: list[str] = [
            FORMATS[get_key(aformat, FORMATS)]
            for aformat in self.opt_manager.options.get("selected_audio_formats", [])
        ]

        if vformats:
            lb_headers.add_header(_("Video"))
            lb_headers.add_items(vformats)

        if aformats:
            lb_headers.add_header(_("Audio"))
            lb_headers.add_items(aformats)

        current_index = lb_headers.FindString(
            FORMATS[self.opt_manager.options.get("selected_format", "")]
        )

        if current_index == wx.NOT_FOUND:
            lb_headers.SetSelection(0)
        else:
            lb_headers.SetSelection(current_index)

        self._popup_ctrl.value = lb_headers.GetSelection()

        self._update_videoformat(None)

    # noinspection PyUnusedLocal
    def _update_videoformat(self, event):
        lb_headers = self._get_listbox_headers()
        selection = lb_headers.GetStringSelection()
        selected_format = get_key(selection, FORMATS, "0")
        self.opt_manager.options["selected_format"] = selected_format

        if selected_format in VIDEO_FORMATS:
            self.opt_manager.options["video_format"] = selected_format
            self.opt_manager.options[
                "audio_format"
            ] = ""  # NOTE Set to default value, check parsers.py
        elif selected_format in AUDIO_FORMATS:
            self.opt_manager.options["video_format"] = "0"
            self.opt_manager.options["audio_format"] = selected_format
        else:
            self.opt_manager.options["video_format"] = "0"
            self.opt_manager.options["audio_format"] = ""

        self._videoformat_combobox.SetText(FORMATS[selected_format])

    # noinspection PyUnusedLocal
    def _update_savepath(self, event):
        try:
            path: Path = Path(self._path_combobox.GetValue()).resolve()
            if not path.exists():
                os.makedirs(str(path))
        except OSError:
            # Avoid [WinError 433] Driver removed ?!
            path: Path = Path().home() / Path("Videos")
            self._path_combobox.SetValue(str(path))

        self.opt_manager.options["save_path"] = str(path)

    # noinspection PyUnusedLocal
    def _on_delete(self, event):
        index = self._status_list.get_next_selected()

        if index == -1:
            dlg = ButtonsChoiceDialog(
                self,
                [_("Remove all"), _("Remove completed")],
                _("No items selected. Please pick an action"),
                _("Delete"),
                self._dark_mode,
            )

            ret_code = dlg.ShowModal()
            dlg.Destroy()

            # REFACTOR Maybe add this functionality directly to DownloadList?
            if ret_code == 1:
                for ditem in self._download_list.get_items():
                    if ditem.stage != "Active":
                        self._status_list.remove_row(
                            self._download_list.index(ditem.object_id)
                        )
                        self._download_list.remove(ditem.object_id)

            if ret_code == 2:
                for ditem in self._download_list.get_items():
                    if ditem.stage == "Completed":
                        self._status_list.remove_row(
                            self._download_list.index(ditem.object_id)
                        )
                        self._download_list.remove(ditem.object_id)
        else:
            result = True

            if self.opt_manager.options.get("confirm_deletion", True):
                dlg = MessageDialog(
                    self,
                    _("Are you sure you want to remove selected items?"),
                    _("Delete"),
                    self._dark_mode,
                )

                result = dlg.ShowModal() == wx.ID_YES
                dlg.Destroy()

            if result:
                while index >= 0:
                    object_id = self._status_list.GetItemData(index)
                    selected_download_item = self._download_list.get_item(object_id)

                    if selected_download_item.stage == "Active":
                        self._create_popup(
                            _("Item is active, cannot remove"),
                            self.WARNING_LABEL,
                            wx.OK | wx.ICON_EXCLAMATION,
                        )
                    else:
                        self._status_list.remove_row(index)
                        self._download_list.remove(object_id)
                        index -= 1

                    index = self._status_list.get_next_selected(index)

        self._update_pause_button(None)

    # noinspection PyUnusedLocal
    def _on_play(self, event):
        selected_rows = self._status_list.get_all_selected()

        if selected_rows:
            for selected_row in selected_rows:
                object_id = self._status_list.GetItemData(selected_row)
                selected_download_item = self._download_list.get_item(object_id)

                if selected_download_item.stage == "Completed":
                    if selected_download_item.filenames:
                        filename = selected_download_item.get_files()[-1]
                        open_file(filename)
                else:
                    self._create_popup(
                        _("Item is not completed"),
                        self.INFO_LABEL,
                        wx.OK | wx.ICON_INFORMATION,
                    )

    # noinspection PyUnusedLocal,PyProtectedMember
    def _on_arrow_up(self, event):
        index = self._status_list.get_next_selected()

        if index != -1:
            while index >= 0:
                object_id: int | None = self._status_list.GetItemData(index)
                download_item = self._download_list.get_item(object_id)

                assert object_id is not None
                assert download_item is not None

                new_index = index - 1
                new_index = max(new_index, 0)

                if not self._status_list.IsSelected(new_index):
                    self._download_list.move_up(object_id)
                    self._status_list.move_item_up(index)
                    self._status_list._update_from_item(new_index, download_item)

                index = self._status_list.get_next_selected(index)

    # noinspection PyUnusedLocal,PyProtectedMember
    def _on_arrow_down(self, event):
        index = self._status_list.get_next_selected(reverse=True)

        if index != -1:
            while index >= 0:
                object_id: int | None = self._status_list.GetItemData(index)
                download_item = self._download_list.get_item(object_id)

                assert object_id is not None
                assert download_item is not None

                new_index = index + 1
                if new_index >= self._status_list.GetItemCount():
                    new_index = self._status_list.GetItemCount() - 1

                if not self._status_list.IsSelected(new_index):
                    self._download_list.move_down(object_id)
                    self._status_list.move_item_down(index)
                    self._status_list._update_from_item(new_index, download_item)

                index = self._status_list.get_next_selected(index, True)

    # noinspection PyUnusedLocal,PyProtectedMember
    def _on_reload(self, event):
        selected_rows = self._status_list.get_all_selected()

        if not selected_rows:
            for index, download_item in enumerate(self._download_list.get_items()):
                if download_item.stage in ("Paused", "Completed", "Error"):
                    # Store the old savepath because reset is going to remove it
                    savepath = download_item.path
                    download_item.reset()
                    download_item.path = savepath
                    self._status_list._update_from_item(index, download_item)
        else:
            for selected_row in selected_rows:
                object_id: int | None = self._status_list.GetItemData(selected_row)
                download_item = self._download_list.get_item(object_id)

                assert download_item is not None

                if download_item.stage in ("Paused", "Completed", "Error"):
                    # Store the old savepath because reset is going to remove it
                    savepath = download_item.path
                    download_item.reset()
                    download_item.path = savepath
                    self._status_list._update_from_item(selected_row, download_item)

            self._update_pause_button(None)

    # noinspection PyUnusedLocal,PyProtectedMember
    def _on_pause(self, event):
        selected_rows = self._status_list.get_all_selected()

        if selected_rows:
            # REFACTOR Use widgets.DoubleStageButton for this and check stage
            if self._buttons["pause"].GetLabel() == _("Pause"):
                new_state = "Paused"
            else:
                new_state = "Queued"

            for selected_row in selected_rows:
                object_id: int | None = self._status_list.GetItemData(selected_row)
                download_item = self._download_list.get_item(object_id)

                assert object_id is not None
                assert download_item is not None

                if download_item.stage in ["Queued", "Paused"]:
                    self._download_list.change_stage(object_id, new_state)

                self._status_list._update_from_item(selected_row, download_item)

            self._update_pause_button(None)

    # noinspection PyUnusedLocal
    def _on_start(self, event):
        if self.download_manager is None:
            if self.update_thread is not None and self.update_thread.is_alive():
                self._create_popup(
                    _("Update in progress. Please wait for the update to complete"),
                    self.WARNING_LABEL,
                    wx.OK | wx.ICON_EXCLAMATION,
                )
            else:
                self._start_download()
        else:
            self.download_manager.stop_downloads()

    # noinspection PyUnusedLocal
    def _on_savepath(self, event):
        dlg = wx.DirDialog(
            self, self.CHOOSE_DIRECTORY, self._path_combobox.GetStringSelection()
        )

        if dlg.ShowModal() == wx.ID_OK:
            path = dlg.GetPath()

            self._path_combobox.Append(path)
            self._path_combobox.SetValue(path)
            self._update_savepath(None)

        dlg.Destroy()

    # noinspection PyUnusedLocal
    def _on_add(self, event):
        urls = self._get_urls()

        if not urls:
            self._create_popup(
                self.PROVIDE_URL_MSG, self.WARNING_LABEL, wx.OK | wx.ICON_EXCLAMATION
            )
        else:
            self._url_list.Clear()
            options = self._options_parser.parse(self.opt_manager.options)

            for url in urls:
                download_item = DownloadItem(url, options)
                download_item.path = self.opt_manager.options.get("save_path", ".")

                if not self._download_list.has_item(download_item.object_id):
                    self._status_list.bind_item(download_item)
                    self._download_list.insert(download_item)

    def _on_settings(self, event):
        event_object_pos = event.EventObject.GetPosition()
        event_object_height = event.EventObject.GetSize()[1]
        event_object_pos = (
            event_object_pos[0],
            event_object_pos[1] + event_object_height,
        )
        self.PopupMenu(self._settings_menu, event_object_pos)

    # noinspection PyUnusedLocal
    def _on_viewlog(self, event):
        if self.log_manager is None:
            self._create_popup(
                _("Logging is disabled"),
                self.WARNING_LABEL,
                wx.OK | wx.ICON_EXCLAMATION,
            )
        else:
            log_window = LogGUI(self)
            log_window.load(self.log_manager.log_file)
            log_window.Show()

    # noinspection PyUnusedLocal
    def _on_about(self, event):
        info = wx.adv.AboutDialogInfo()

        if self.app_icon is not None:
            info.SetIcon(self.app_icon)

        info.SetName(__appname__)
        info.SetVersion(__version__)
        info.SetDescription(__descriptionfull__)
        info.SetWebSite(__projecturl__)
        info.SetLicense(__licensefull__)
        info.AddDeveloper(
            __author__
            + "\n"
            + __maintainer__
            + "\n"
            + "see AUTHORS file for the complete list."
        )

        wx.adv.AboutBox(info)

    @staticmethod
    def _set_publisher(handler: Callable, topic: str):
        """Sets a handler for the given topic.

        Args:
            handler (function): Can be any function with one parameter
                the message that the caller sends.

            topic (string): Can be any string that identifies the caller.
                You can bind multiple handlers on the same topic or
                multiple topics on the same handler.

        """
        Publisher.subscribe(handler, topic)

    def _create_statictext(self, label):
        return wx.StaticText(self._panel, label=label)

    def _create_bitmap_button(self, icon, size=(-1, -1), handler=None):
        button = wx.BitmapButton(
            self._panel, bitmap=icon, size=size, style=wx.NO_BORDER
        )

        if handler is not None:
            button.Bind(wx.EVT_BUTTON, handler)

        return button

    def _create_static_bitmap(self, icon, event_handler=None):
        static_bitmap = wx.StaticBitmap(self._panel, bitmap=icon)

        if event_handler is not None:
            static_bitmap.Bind(wx.EVT_LEFT_DCLICK, event_handler)

        return static_bitmap

    def _create_textctrl(self, style=None, event_handler=None):
        if style is None:
            textctrl = wx.TextCtrl(self._panel)
        else:
            textctrl = wx.TextCtrl(self._panel, style=style)

        if event_handler is not None:
            textctrl.Bind(wx.EVT_TEXT_PASTE, event_handler)
            textctrl.Bind(wx.EVT_MIDDLE_DOWN, event_handler)

        if os.name == "nt":
            # Enable CTRL+A on Windows
            def win_ctrla_eventhandler(event):
                if event.GetKeyCode() == wx.WXK_CONTROL_A:
                    event.GetEventObject().SelectAll()

                event.Skip()

            textctrl.Bind(wx.EVT_CHAR, win_ctrla_eventhandler)

        return textctrl

    @staticmethod
    def _create_popup(text, title, style):
        wx.MessageBox(text, title, style)

    def _set_layout(self):
        """Sets the layout of the main window."""
        main_sizer = wx.BoxSizer()
        panel_sizer = wx.BoxSizer(wx.VERTICAL)

        top_sizer = wx.BoxSizer(wx.HORIZONTAL)
        top_sizer.Add(self._url_text, 0, wx.ALIGN_BOTTOM | wx.BOTTOM, 5)
        top_sizer.AddStretchSpacer(1)
        top_sizer.Add(self._buttons["settings"])
        panel_sizer.Add(top_sizer, 0, wx.EXPAND | wx.BOTTOM, 5)

        panel_sizer.Add(self._url_list, 1, wx.EXPAND)

        mid_sizer = wx.BoxSizer(wx.HORIZONTAL)
        mid_sizer.Add(self._folder_icon)
        mid_sizer.AddSpacer(3)
        mid_sizer.Add(self._path_combobox, 2, wx.ALIGN_CENTER_VERTICAL)
        mid_sizer.AddSpacer(5)
        mid_sizer.Add(self._buttons["savepath"], flag=wx.ALIGN_CENTER_VERTICAL)
        mid_sizer.AddStretchSpacer(1)
        mid_sizer.Add(self._videoformat_combobox, 0, wx.ALIGN_CENTER_VERTICAL)
        mid_sizer.AddSpacer(5)
        mid_sizer.Add(self._buttons["add"], flag=wx.ALIGN_CENTER_VERTICAL)
        panel_sizer.Add(mid_sizer, 0, wx.EXPAND | wx.TOP | wx.BOTTOM, 10)

        panel_sizer.Add(self._download_text, 0, wx.BOTTOM, 5)
        panel_sizer.Add(self._status_list, 2, wx.EXPAND)

        bottom_sizer = wx.BoxSizer(wx.HORIZONTAL)
        bottom_sizer.Add(self._buttons["delete"])
        bottom_sizer.AddSpacer(5)
        bottom_sizer.Add(self._buttons["play"])
        bottom_sizer.AddSpacer(5)
        bottom_sizer.Add(self._buttons["up"])
        bottom_sizer.AddSpacer(5)
        bottom_sizer.Add(self._buttons["down"])
        bottom_sizer.AddSpacer(5)
        bottom_sizer.Add(self._buttons["reload"])
        bottom_sizer.AddSpacer(5)
        bottom_sizer.Add(self._buttons["pause"])
        bottom_sizer.AddStretchSpacer(1)
        bottom_sizer.Add(self._buttons["start"])
        panel_sizer.Add(bottom_sizer, 0, wx.EXPAND | wx.TOP, 5)

        main_sizer.Add(panel_sizer, 1, wx.ALL | wx.EXPAND, 10)

        self._panel.SetSizer(main_sizer)
        self._panel.Layout()

    def _update_youtubedl(self):
        """Update youtube-dl binary to the latest version."""
        if self.download_manager is not None and self.download_manager.is_alive():
            self._create_popup(
                self.DOWNLOAD_ACTIVE, self.WARNING_LABEL, wx.OK | wx.ICON_EXCLAMATION
            )
        elif self.update_thread is not None and self.update_thread.is_alive():
            self._create_popup(
                self.UPDATE_ACTIVE, self.INFO_LABEL, wx.OK | wx.ICON_INFORMATION
            )
        else:
            self.update_thread = UpdateThread(self.opt_manager)

    def _status_bar_write(self, msg: str):
        """Display msg in the status bar."""
        self._status_bar.SetStatusText(msg)

    def _reset_widgets(self):
        """Resets GUI widgets after update or download process."""
        self._buttons["start"].SetLabel(self.START_LABEL)
        self._buttons["start"].SetToolTip(wx.ToolTip(self.START_LABEL))
        self._buttons["start"].SetBitmap(self._bitmaps["start"], wx.TOP)

    def _print_stats(self):
        """Display download stats in the status bar."""
        suc_downloads = self.download_manager.successful
        dtime = get_time(self.download_manager.time_it_took)

        msg = self.SUCC_REPORT_MSG.format(
            suc_downloads,
            dtime["days"],
            dtime["hours"],
            dtime["minutes"],
            dtime["seconds"],
        )

        self._status_bar_write(msg)

    def _after_download(self):
        """Run tasks after download process has been completed.

        Note:
            Here you can add any tasks you want to run after the
            download process has been completed.

        """
        if self.opt_manager.options.get("shutdown", False):
            dlg = ShutdownDialog(
                self, 60, _("Shutting down in {0} second(s)"), _("Shutdown")
            )
            result = dlg.ShowModal() == wx.ID_OK
            dlg.Destroy()

            if result:
                self.opt_manager.save_to_file()
                success = shutdown_sys(
                    self.opt_manager.options.get("sudo_password", "")
                )

                if success:
                    self._status_bar_write(self.SHUTDOWN_MSG)
                else:
                    self._status_bar_write(self.SHUTDOWN_ERR)
        else:
            if self.opt_manager.options["show_completion_popup"]:
                self._create_popup(
                    self.DL_COMPLETED_MSG, self.INFO_LABEL, wx.OK | wx.ICON_INFORMATION
                )

    # noinspection PyUnusedLocal,PyProtectedMember
    def _download_worker_handler(self, signal: str, data: dict[str, Any] | None = None):
        """downloadmanager.Worker thread handler.

        Handles messages from the Worker thread.

        Args:
            See downloadmanager.Worker _talk_to_gui() method.

        """

        download_item: DownloadItem | None = self._download_list.get_item(data["index"])

        assert download_item is not None
        assert data is not None

        download_item.update_stats(data)
        row = self._download_list.index(data["index"])
        self._status_list._update_from_item(row, download_item)

    # noinspection PyUnusedLocal
    def _download_manager_handler(
        self, signal: str, data: dict[str, str] | None = None
    ):
        """downloadmanager.DownloadManager thread handler.

        Handles messages from the DownloadManager thread.

        Args:
            See downloadmanager.DownloadManager _talk_to_gui() method.

        """
        # TODO: REFACTOR Manage better the signal stage

        if signal == "finished":
            self._print_stats()
            self._reset_widgets()
            self.download_manager = None
            self._app_timer.Stop()
            self._after_download()
        elif signal == "closed":
            self._status_bar_write(self.CLOSED_MSG)
            self._reset_widgets()
            self.download_manager = None
            self._app_timer.Stop()
        elif signal == "closing":
            self._status_bar_write(self.CLOSING_MSG)
            # NOTE Remove from here and downloadmanager
            # since now we have the wx.Timer to check progress

    def _update_handler(self, signal: str, data: list[str] = None):
        """updatemanager.UpdateThread thread handler.

        Handles messages from the UpdateThread thread.

        Args:
            See updatemanager.UpdateThread _talk_to_gui() method.

        """

        if signal == "download":
            self._status_bar_write(self.UPDATING_MSG)
        elif signal == "error":
            self._status_bar_write(self.UPDATE_ERR_MSG.format(data))
        elif signal == "correct":
            self._status_bar_write(self.UPDATE_SUCC_MSG)
        else:
            self._reset_widgets()
            self.update_thread = None

    def _get_urls(self) -> list[str]:
        """Returns urls list."""
        return [line for line in self._url_list.GetValue().split("\n") if line]

    def _start_download(self):
        if self._status_list.is_empty():
            self._create_popup(
                _("No items to download"),
                self.WARNING_LABEL,
                wx.OK | wx.ICON_EXCLAMATION,
            )
        else:
            self._app_timer.Start(100)
            self.download_manager = DownloadManager(
                self, self._download_list, self.opt_manager, self.log_manager
            )

            self._status_bar_write(self.DOWNLOAD_STARTED)
            self._buttons["start"].SetLabel(self.STOP_LABEL)
            self._buttons["start"].SetToolTip(wx.ToolTip(self.STOP_LABEL))
            self._buttons["start"].SetBitmap(self._bitmaps["stop"], wx.TOP)

    def _paste_from_clipboard(self):
        """Paste the content of the clipboard to the self._url_list widget.
        It also adds a new line at the end of the data if not exist.

        """
        if wx.TheClipboard.IsOpened():
            return

        if wx.TheClipboard.Open():
            if wx.TheClipboard.IsSupported(wx.DataFormat(wx.DF_TEXT)):

                data = wx.TextDataObject()
                wx.TheClipboard.GetData(data)

                data = data.GetText()

                if data != "" and data[-1] != "\n":
                    data += "\n"

                self._url_list.WriteText(data)

            wx.TheClipboard.Close()

    def _on_urllist_edit(self, event):
        """Event handler of the self._url_list widget.

        This method is triggered when the users pastes text into
        the URLs list either by using CTRL+V or by using the middle
        click of the mouse.

        """
        if event.GetEventType() == wx.EVT_TEXT_PASTE.typeId:
            self._paste_from_clipboard()
        else:
            wx.TheClipboard.UsePrimarySelection(True)
            self._paste_from_clipboard()
            wx.TheClipboard.UsePrimarySelection(False)

    # noinspection PyUnusedLocal
    def _on_update(self, event):
        """Event handler of the self._update_btn widget.

        This method is used when the update button is pressed to start
        the update process.

        Note:
            Currently there is not way to stop the update process.

        """
        if self.opt_manager.options.get("disable_update", False):
            self._create_popup(
                _(
                    "Updates are disabled for your system. "
                    "Please use the system's package manager to update the CLI Backend."
                ),
                self.INFO_LABEL,
                wx.OK | wx.ICON_INFORMATION,
            )
        else:
            self._update_youtubedl()

    # noinspection PyUnusedLocal
    def _on_options(self, event):
        """Event handler of the self._options_btn widget.

        This method is used when the options button is pressed to show
        the options window.

        """
        self._options_frame.load_all_options()
        self._options_frame.Show()

    # noinspection PyUnusedLocal
    def _on_close(self, event):
        """Event handler for the wx.EVT_CLOSE event.

        This method is used when the user tries to close the program
        to save the options and make sure that the download & update
        processes are not running.

        """
        result = True

        if self.opt_manager.options.get("confirm_exit", True):
            dlg = MessageDialog(
                self,
                _("Are you sure you want to exit?"),
                _("Exit"),
                self._dark_mode,
            )

            result = dlg.ShowModal() == wx.ID_YES
            dlg.Destroy()

        if result:
            self.close()

    def close(self):
        if self.download_manager:
            self.download_manager.stop_downloads()
            self.download_manager.join()

        if self.update_thread:
            self.update_thread.join()

        # Store main-options frame size
        self.opt_manager.options["main_win_size"] = self.GetSize()
        self.opt_manager.options["opts_win_size"] = self._options_frame.GetSize()

        self.opt_manager.options["save_path_dirs"] = self._path_combobox.GetStrings()

        self._options_frame.save_all_options()
        self.opt_manager.save_to_file()

        self.Destroy()



"""
This will generate the .pot and .mo files for the application domain and
languages defined below.

The .po and .mo files are placed as per convention in

"__packagename__/locale/lang/LC_MESSAGES"

The .pot file is placed in the locale folder.

This script or something similar should be added to your build process.

The actual translation work is normally done using a tool like poEdit or
similar, it allows you to generate a particular language catalog from the .pot
file or to use the .pot to merge new translations into an existing language
catalog.

"""


import subprocess
import sys
from pathlib import Path

import wx

__packagename__ = "youtube_dl_gui"

# language domain
langDomain = __packagename__
# languages you want to support
supLang = {
    "en_US": wx.LANGUAGE_ENGLISH,
    "es_CU": wx.LANGUAGE_SPANISH,
}

# we remove English as source code strings are in English
supportedLang = [lang for lang in supLang if lang != "en_US"]

PATH = Path(__file__).resolve().parent
appFolder = PATH.parent.joinpath(__packagename__)

# setup some stuff to get at Python I18N tools/utilities

pyPath = Path(sys.executable).resolve()
pyExe = pyPath.name
pyFolder = pyPath.parent
pyToolsFolder = pyFolder.joinpath("Tools")
pyI18nFolder = pyToolsFolder.joinpath("i18n")
pyGettext = pyI18nFolder.joinpath("pygettext.py")
pyMsgfmt = pyI18nFolder.joinpath("msgfmt.py")
outFolder = appFolder.joinpath("locale")

# build command for pygettext
gtOptions = f"-a -d {langDomain} -o {langDomain}.pot -p {outFolder} {appFolder}"
tCmd = f"{pyExe} {pyGettext} {gtOptions}"

print("Generating the .pot file")
print(f"cmd: {tCmd}")
rCode = subprocess.call(tCmd)

if rCode != 0:
    sys.exit(f"return code: {rCode}")

print("")

for tLang in supportedLang:
    # build command for msgfmt
    langDir = appFolder.joinpath(f"locale/{tLang}/LC_MESSAGES")
    poFile = langDir.joinpath(langDomain).with_suffix(".po")
    tCmd = f"{pyExe} {pyMsgfmt} {poFile}"

    print("Generating the .mo file")
    print(f"cmd: {tCmd}")
    rCode = subprocess.call(tCmd)
    if rCode != 0:
        sys.exit(f"return code: {rCode}")



# type: ignore[misc]
"""yt-dlg module responsible for the options window. """
from __future__ import annotations

import os
from typing import TYPE_CHECKING, Callable, cast

import wx
import wx.adv

from .darktheme import dark_mode
from .flagart import catalog
from .formats import AUDIO_FORMATS, OUTPUT_FORMATS, VIDEO_FORMATS
from .info import __appname__
from .utils import IS_WINDOWS, YOUTUBEDL_BIN, YTDLP_BIN, get_key
from .widgets import LogGUI

if TYPE_CHECKING:
    from .mainframe import MainFrame

_: Callable[[str], str] = wx.GetTranslation
# REFACTOR Move all formats, etc to formats.py


class OptionsFrame(wx.Frame):
    """yt-dlg options frame class.

    Args:
        parent (MainFrame): Parent class.

    """

    FRAMES_MIN_SIZE: tuple[int, int] = (500, 470)

    def __init__(self, parent: MainFrame, darkmode: bool = False):
        wx.Frame.__init__(
            self,
            parent,
            title=_("Options"),
            size=parent.opt_manager.options["opts_win_size"],
        )
        self.parent = parent
        self.opt_manager = self.parent.opt_manager
        self.log_manager = self.parent.log_manager
        self.app_icon: wx.Icon | None = self.parent.app_icon

        if self.app_icon:
            self.SetIcon(self.app_icon)

        self.__dark_mode: bool = darkmode
        self._was_shown: bool = False

        # Create options frame basic components
        self.panel = wx.Panel(self)

        self.notebook = wx.Notebook(self.panel)
        self.separator_line = wx.StaticLine(self.panel)
        self.reset_button = wx.Button(self.panel, label=_("Reset"))
        self.close_button = wx.Button(self.panel, label=_("Close"))

        # Create tabs
        tab_args: tuple[OptionsFrame, wx.Notebook] = (self, self.notebook)

        self.tabs: tuple[
            tuple[GeneralTab, str],
            tuple[FormatsTab, str],
            tuple[DownloadsTab, str],
            tuple[AdvancedTab, str],
            tuple[ExtraTab, str],
        ] = (
            (GeneralTab(*tab_args), _("General")),
            (FormatsTab(*tab_args), _("Formats")),
            (DownloadsTab(*tab_args), _("Downloads")),
            (AdvancedTab(*tab_args), _("Advanced")),
            (ExtraTab(*tab_args), _("Extra")),
        )

        # Add tabs on notebook
        for tab, label in self.tabs:
            self.notebook.AddPage(tab, label)

        # Bind events
        self.Bind(wx.EVT_BUTTON, self._on_reset, self.reset_button)
        self.Bind(wx.EVT_BUTTON, self._on_close, self.close_button)
        self.Bind(wx.EVT_CLOSE, self._on_close)

        self.SetMinSize(self.FRAMES_MIN_SIZE)

        self._set_layout()
        # Set Dar Theme for Notebook
        dark_mode(self.notebook, self.__dark_mode)
        self.load_all_options()

    def _set_layout(self) -> None:
        main_sizer = wx.BoxSizer(wx.VERTICAL)

        main_sizer.Add(self.notebook, 1, wx.EXPAND | wx.ALL, border=5)
        main_sizer.Add(self.separator_line, 0, wx.EXPAND)

        buttons_sizer = wx.BoxSizer(wx.HORIZONTAL)
        buttons_sizer.Add(self.reset_button)
        buttons_sizer.AddSpacer(5)
        buttons_sizer.Add(self.close_button)

        main_sizer.Add(buttons_sizer, flag=wx.ALIGN_RIGHT | wx.ALL, border=5)

        self.panel.SetSizer(main_sizer)

        self.panel.Layout()

    # noinspection PyProtectedMember,PyUnusedLocal
    def _on_close(self, event) -> None:
        """Event handler for wx.EVT_CLOSE event."""
        self.save_all_options()
        # REFACTOR Parent create specific callback
        self.parent._update_videoformat_combobox()
        self.Hide()

    # noinspection PyUnusedLocal
    def _on_reset(self, event) -> None:
        """Event handler for the reset button wx.EVT_BUTTON event."""
        self.reset()
        self.parent.reset()

    def reset(self) -> None:
        """Reset the default options."""
        self.opt_manager.load_default()
        self.load_all_options()

    def load_all_options(self) -> None:
        """Load all the options on each tab."""
        for tab, _label in self.tabs:
            cast("TabPanel", tab).load_options()

    def save_all_options(self) -> None:
        """Save all the options from all the tabs back to the OptionsManager."""
        for tab, _label in self.tabs:
            cast("TabPanel", tab).save_options()

    def Show(self, *args, **kwargs) -> None:
        """Shows options frame centered"""
        # CenterOnParent can't go to main frame's __init__ as main frame may change
        # own position and options frame won't be centered on main frame anymore.
        if not self._was_shown:
            self._was_shown = True
            self.CenterOnParent()
        return wx.Frame.Show(self, *args, **kwargs)

    def is_dark(self) -> bool:
        return self.__dark_mode


class TabPanel(wx.Panel):
    """Main tab class from which each tab inherits.

    Args:
        parent (OptionsFrame): The parent of all tabs.

        notebook (wx.Notebook): The container for each tab.

    Notes:
        In order to use a different size you must overwrite the below *_SIZE
        attributes on the corresponding child object.

    """

    CHECKBOX_SIZE: tuple[int, int] = wx.DefaultSize
    if IS_WINDOWS:
        # Make checkboxes look the same on Windows
        CHECKBOX_SIZE = (-1, 20)

    BUTTONS_SIZE: tuple[int, int] = wx.DefaultSize
    TEXTCTRL_SIZE: tuple[int, int] = wx.DefaultSize
    SPINCTRL_SIZE: tuple[int, int] = wx.DefaultSize

    CHECKLISTBOX_SIZE: tuple[int, int] = (-1, 80)
    LISTBOX_SIZE: tuple[int, int] = (-1, 80)

    def __init__(self, parent: OptionsFrame, notebook: wx.Notebook):
        super().__init__(notebook)
        # REFACTOR Maybe add methods to access those
        # save_options(key, value)
        # load_options(key)
        self.opt_manager = parent.opt_manager
        self.log_manager = parent.log_manager
        self.app_icon = parent.app_icon

        self.reset_handler = parent.reset

    def load_options(self):
        return NotImplemented

    def save_options(self):
        return NotImplemented

    # Shortcut methods below

    def crt_button(
        self, label: str, event_handler: Callable | None = None
    ) -> wx.Button:
        button = wx.Button(self, label=label, size=self.BUTTONS_SIZE)

        if event_handler:
            button.Bind(wx.EVT_BUTTON, event_handler)

        return button

    def crt_checkbox(
        self, label: str, event_handler: Callable | None = None
    ) -> wx.CheckBox:
        checkbox = wx.CheckBox(self, label=label, size=self.CHECKBOX_SIZE)

        if event_handler:
            checkbox.Bind(wx.EVT_CHECKBOX, event_handler)

        return checkbox

    def crt_textctrl(self, style: int | None = None) -> wx.TextCtrl:
        if style is not None:
            return wx.TextCtrl(self, size=self.TEXTCTRL_SIZE, style=style)

        return wx.TextCtrl(self, size=self.TEXTCTRL_SIZE)

    def crt_combobox(
        self, choices, size=(-1, -1), event_handler: Callable | None = None
    ) -> wx.ComboBox:
        combobox = wx.ComboBox(self, choices=choices, size=size, style=wx.CB_READONLY)

        if event_handler:
            combobox.Bind(wx.EVT_COMBOBOX, event_handler)

        return combobox

    def crt_bitmap_combobox(
        self,
        choices: list[tuple[str, str]],
        size: tuple[int, int] = (-1, -1),
        event_handler: Callable | None = None,
    ) -> wx.adv.BitmapComboBox:
        combobox = wx.adv.BitmapComboBox(self, size=size, style=wx.CB_READONLY)

        for item in choices:
            lang_code, lang_name = item
            _lang, country = lang_code.split("_")

            if country in catalog:
                flag_bmp: wx.Bitmap = catalog[country].GetBitmap()
            else:
                flag_bmp = catalog["BLANK"].GetBitmap()

            combobox.Append(lang_name, flag_bmp)

        if event_handler:
            combobox.Bind(wx.EVT_COMBOBOX, event_handler)

        return combobox

    def crt_spinctrl(
        self,
        spin_range: tuple[int, int] = (0, 9999),
        size: tuple[int, int] | None = None,
    ) -> wx.SpinCtrl:
        if not size:
            size = (130, -1) if not IS_WINDOWS else self.SPINCTRL_SIZE
        spinctrl = wx.SpinCtrl(self, size=size)
        spinctrl.SetRange(*spin_range)

        return spinctrl

    def crt_statictext(self, label: str) -> wx.StaticText:
        return wx.StaticText(self, wx.ID_ANY, label)

    def crt_staticbox(self, label: str) -> wx.StaticBox:
        return wx.StaticBox(self, wx.ID_ANY, label)

    def crt_checklistbox(
        self, choices: list[str] | None, style: int | None = None
    ) -> wx.CheckListBox:
        if style is not None:
            return wx.CheckListBox(
                self, choices=choices, style=style, size=self.CHECKLISTBOX_SIZE
            )

        return wx.CheckListBox(self, choices=choices, size=self.CHECKLISTBOX_SIZE)

    def crt_listbox(
        self, choices: list[str] | None, style: int | None = None
    ) -> wx.ListBox:
        if style is not None:
            return wx.ListBox(
                self, choices=choices, style=style, size=self.LISTBOX_SIZE
            )

        return wx.ListBox(self, choices=choices, size=self.LISTBOX_SIZE)


class GeneralTab(TabPanel):

    OUTPUT_TEMPLATES: list[str] = [
        "Id",
        "Title",
        "Ext",
        "Uploader",
        "Resolution",
        "Autonumber",
        "",
        "View Count",
        "Like Count",
        "Dislike Count",
        "Comment Count",
        "Average Rating",
        "Age Limit",
        "Width",
        "Height",
        "Extractor",
        "",
        "Playlist",
        "Playlist Index",
    ]

    BUTTONS_SIZE: tuple[int, int] = (35, -1)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        _underscore: Callable = _

        for key, value in OUTPUT_FORMATS.items():
            OUTPUT_FORMATS[key] = _underscore(value)

        # Lang code = <ISO 639-1>_<ISO 3166-1 alpha-2>
        self.LOCALE_NAMES: dict[str, str] = {
            "sq_AL": _("Albanian"),
            "ar_SA": _("Arabic"),
            "es_CU": _("Cuba"),
            "cs_CZ": _("Czech"),
            "de_DE": _("German"),
            "en_US": _("English"),
            "fr_FR": _("French"),
            "it_IT": _("Italian"),
            "ja_JP": _("Japanese"),
            "ko_KR": _("Korean"),
            "pl_PL": _("Polish"),
            "pt_BR": _("Portuguese"),
            "ru_RU": _("Russian"),
            "sk_SK": _("Slovak"),
            "es_ES": _("Spanish"),
            "zh_CN": _("Simplified Chinese"),
            "zh_TW": _("Traditional Chinese"),
        }

        self.language_label = self.crt_statictext(_("Language"))
        self.language_combobox = self.crt_bitmap_combobox(
            list(self.LOCALE_NAMES.items()), event_handler=self._on_restart
        )

        self.filename_format_label = self.crt_statictext(_("Filename format"))
        self.filename_format_combobox = self.crt_combobox(
            list(OUTPUT_FORMATS.values()), event_handler=self._on_filename
        )
        self.filename_custom_format = self.crt_textctrl()
        self.filename_custom_format_button = self.crt_button("...", self._on_format)

        self.filename_opts_label = self.crt_statictext(_("Filename options"))
        self.filename_ascii_checkbox = self.crt_checkbox(
            _("Restrict filenames to ASCII")
        )

        self.more_opts_label = self.crt_statictext(_("More options"))
        self.dark_mode_checkbox = self.crt_checkbox(
            _("Dark theme"), event_handler=self._on_restart
        )
        self.confirm_exit_checkbox = self.crt_checkbox(_("Confirm on exit"))
        self.confirm_deletion_checkbox = self.crt_checkbox(_("Confirm item deletion"))
        self.show_completion_popup_checkbox = self.crt_checkbox(
            _("Inform me on download completion")
        )

        self.shutdown_checkbox = self.crt_checkbox(
            _("Shutdown on download completion"), event_handler=self._on_shutdown
        )
        self.sudo_textctrl = self.crt_textctrl(wx.TE_PASSWORD)

        # Build the menu for the custom format button
        self.custom_format_menu = self._build_custom_format_menu()

        self._set_layout()

        if IS_WINDOWS:
            self.sudo_textctrl.Hide()

        self.sudo_textctrl.SetToolTip(wx.ToolTip(_("SUDO password")))

    def _set_layout(self) -> None:
        main_sizer = wx.BoxSizer(wx.HORIZONTAL)
        vertical_sizer = wx.BoxSizer(wx.VERTICAL)

        vertical_sizer.Add(self.language_label)
        vertical_sizer.Add(self.language_combobox, flag=wx.EXPAND | wx.ALL, border=5)

        vertical_sizer.Add(self.filename_format_label, flag=wx.TOP, border=5)
        vertical_sizer.Add(
            self.filename_format_combobox, flag=wx.EXPAND | wx.ALL, border=5
        )

        custom_format_sizer = wx.BoxSizer(wx.HORIZONTAL)
        custom_format_sizer.Add(
            self.filename_custom_format, 1, wx.ALIGN_CENTER_VERTICAL
        )
        custom_format_sizer.AddSpacer(5)
        custom_format_sizer.Add(self.filename_custom_format_button)

        vertical_sizer.Add(
            custom_format_sizer,
            flag=wx.EXPAND | wx.LEFT | wx.RIGHT | wx.BOTTOM,
            border=5,
        )

        vertical_sizer.Add(self.filename_opts_label, flag=wx.TOP, border=5)
        vertical_sizer.Add(self.filename_ascii_checkbox, flag=wx.ALL, border=5)

        vertical_sizer.Add(self.more_opts_label, flag=wx.TOP, border=5)
        vertical_sizer.Add(self.dark_mode_checkbox, flag=wx.ALL, border=5)
        vertical_sizer.Add(
            self.confirm_exit_checkbox,
            flag=wx.LEFT | wx.RIGHT | wx.BOTTOM,
            border=5,
        )
        vertical_sizer.Add(
            self.confirm_deletion_checkbox,
            flag=wx.LEFT | wx.RIGHT | wx.BOTTOM,
            border=5,
        )
        vertical_sizer.Add(
            self.show_completion_popup_checkbox,
            flag=wx.LEFT | wx.RIGHT | wx.BOTTOM,
            border=5,
        )

        shutdown_sizer = wx.BoxSizer(wx.HORIZONTAL)
        shutdown_sizer.Add(self.shutdown_checkbox)
        shutdown_sizer.AddSpacer(-1)
        shutdown_sizer.Add(self.sudo_textctrl, 1)

        vertical_sizer.Add(
            shutdown_sizer, flag=wx.EXPAND | wx.LEFT | wx.RIGHT | wx.BOTTOM, border=5
        )

        main_sizer.Add(vertical_sizer, 1, wx.EXPAND | wx.ALL, border=5)
        self.SetSizer(main_sizer)

    def _build_custom_format_menu(self) -> wx.Menu:
        menu = wx.Menu()

        for template in self.OUTPUT_TEMPLATES:
            if template:
                menu_item = menu.Append(wx.ID_ANY, template)
                menu.Bind(wx.EVT_MENU, self._on_template, menu_item)
            else:
                menu.AppendSeparator()

        return menu

    def _on_template(self, event) -> None:
        """Event handler for the wx.EVT_MENU of the custom_format_menu menu items."""
        label: str = self.custom_format_menu.GetLabelText(event.GetId())
        label = label.lower().replace(" ", "_")

        custom_format: str = self.filename_custom_format.GetValue()

        prefix: str = "." if label == "ext" else "-"
        if not custom_format or custom_format[-1] == os.sep:
            # If the custom format is empty or ends with path separator
            # remove the prefix
            prefix = ""

        template: str = f"{prefix}%({label})s"
        self.filename_custom_format.SetValue(custom_format + template)

    def _on_format(self, event) -> None:
        """Event handler for the wx.EVT_BUTTON of the filename_custom_format_button."""
        event_object_pos = event.EventObject.GetPosition()
        event_object_height = event.EventObject.GetSize()[1]
        event_object_pos = (
            event_object_pos[0],
            event_object_pos[1] + event_object_height,
        )
        self.PopupMenu(self.custom_format_menu, event_object_pos)

    # noinspection PyUnusedLocal
    def _on_restart(self, event) -> None:
        """Event handler for the wx.EVT_COMBOBOX of the language_combobox."""
        wx.MessageBox(
            _("In order for the changes to take effect please restart {0}").format(
                __appname__
            ),
            _("Restart"),
            wx.OK | wx.ICON_INFORMATION,
            self,
        )

    # noinspection PyUnusedLocal
    def _on_filename(self, event) -> None:
        """Event handler for the wx.EVT_COMBOBOX of the filename_format_combobox."""
        custom_selected: bool = (
            self.filename_format_combobox.GetValue() == OUTPUT_FORMATS["3"]
        )

        self.filename_custom_format.Enable(custom_selected)
        self.filename_custom_format_button.Enable(custom_selected)

    # noinspection PyUnusedLocal
    def _on_shutdown(self, event) -> None:
        """Event handler for the wx.EVT_CHECKBOX of the shutdown_checkbox."""
        self.sudo_textctrl.Enable(self.shutdown_checkbox.GetValue())

    def load_options(self) -> None:
        self.language_combobox.SetValue(
            self.LOCALE_NAMES.get(self.opt_manager.options["locale_name"], _("English"))
        )
        self.filename_format_combobox.SetValue(
            OUTPUT_FORMATS[self.opt_manager.options["output_format"]]
        )
        self.filename_custom_format.SetValue(
            self.opt_manager.options["output_template"]
        )
        self.filename_ascii_checkbox.SetValue(
            self.opt_manager.options["restrict_filenames"]
        )
        self.dark_mode_checkbox.SetValue(self.opt_manager.options["dark_mode"])
        self.shutdown_checkbox.SetValue(self.opt_manager.options["shutdown"])
        self.sudo_textctrl.SetValue(self.opt_manager.options["sudo_password"])
        self.confirm_exit_checkbox.SetValue(self.opt_manager.options["confirm_exit"])
        self.show_completion_popup_checkbox.SetValue(
            self.opt_manager.options["show_completion_popup"]
        )
        self.confirm_deletion_checkbox.SetValue(
            self.opt_manager.options["confirm_deletion"]
        )

        # REFACTOR Automatically call on the new methods
        # save_options
        # load_options
        # NOTE Maybe on init add callback?
        self._on_filename(None)
        self._on_shutdown(None)

    def save_options(self) -> None:
        self.opt_manager.options["locale_name"] = get_key(
            self.language_combobox.GetValue(), self.LOCALE_NAMES, "en_US"
        )
        self.opt_manager.options["output_format"] = get_key(
            self.filename_format_combobox.GetValue(), OUTPUT_FORMATS, "1"
        )
        self.opt_manager.options[
            "output_template"
        ] = self.filename_custom_format.GetValue()
        self.opt_manager.options[
            "restrict_filenames"
        ] = self.filename_ascii_checkbox.GetValue()
        self.opt_manager.options["dark_mode"] = self.dark_mode_checkbox.GetValue()
        self.opt_manager.options["shutdown"] = self.shutdown_checkbox.GetValue()
        self.opt_manager.options["sudo_password"] = self.sudo_textctrl.GetValue()
        self.opt_manager.options["confirm_exit"] = self.confirm_exit_checkbox.GetValue()
        self.opt_manager.options[
            "show_completion_popup"
        ] = self.show_completion_popup_checkbox.GetValue()
        self.opt_manager.options[
            "confirm_deletion"
        ] = self.confirm_deletion_checkbox.GetValue()


class FormatsTab(TabPanel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.AUDIO_QUALITY: dict[str, str] = {
            "0": _("high"),
            "5": _("mid"),
            "9": _("low"),
        }

        self.video_formats_label = self.crt_statictext(_("Video formats"))
        self.video_formats_checklistbox = self.crt_checklistbox(
            list(VIDEO_FORMATS.values())
        )

        self.audio_formats_label = self.crt_statictext(_("Audio formats"))
        self.audio_formats_checklistbox = self.crt_checklistbox(
            list(AUDIO_FORMATS.values())
        )

        self.post_proc_opts_label = self.crt_statictext(_("Post-Process options"))
        self.keep_video_checkbox = self.crt_checkbox(_("Keep original files"))
        self.extract_audio_checkbox = self.crt_checkbox(
            _("Extract audio from video file")
        )
        self.embed_thumbnail_checkbox = self.crt_checkbox(
            _("Embed thumbnail in audio file")
        )
        self.add_metadata_checkbox = self.crt_checkbox(_("Add metadata to file"))

        self.audio_quality_label = self.crt_statictext(_("Audio quality"))
        self.audio_quality_combobox = self.crt_combobox(
            list(self.AUDIO_QUALITY.values())
        )

        self._set_layout()

    def _set_layout(self) -> None:
        main_sizer = wx.BoxSizer(wx.HORIZONTAL)
        vertical_sizer = wx.BoxSizer(wx.VERTICAL)

        vertical_sizer.Add(self.video_formats_label)
        vertical_sizer.Add(
            self.video_formats_checklistbox, 1, wx.EXPAND | wx.ALL, border=5
        )

        vertical_sizer.Add(self.audio_formats_label, flag=wx.TOP, border=5)
        vertical_sizer.Add(
            self.audio_formats_checklistbox, 1, wx.EXPAND | wx.ALL, border=5
        )

        vertical_sizer.Add(self.post_proc_opts_label, flag=wx.TOP, border=5)
        vertical_sizer.Add(self.keep_video_checkbox, flag=wx.ALL, border=5)
        vertical_sizer.Add(
            self.extract_audio_checkbox, flag=wx.LEFT | wx.RIGHT | wx.BOTTOM, border=5
        )
        vertical_sizer.Add(
            self.embed_thumbnail_checkbox, flag=wx.LEFT | wx.RIGHT | wx.BOTTOM, border=5
        )
        vertical_sizer.Add(
            self.add_metadata_checkbox, flag=wx.LEFT | wx.RIGHT | wx.BOTTOM, border=5
        )

        audio_quality_sizer = wx.BoxSizer(wx.HORIZONTAL)
        audio_quality_sizer.Add(self.audio_quality_label, flag=wx.ALIGN_CENTER_VERTICAL)
        audio_quality_sizer.AddSpacer(20)
        audio_quality_sizer.Add(self.audio_quality_combobox)

        vertical_sizer.Add(
            audio_quality_sizer, flag=wx.LEFT | wx.RIGHT | wx.BOTTOM, border=5
        )

        main_sizer.Add(vertical_sizer, 1, wx.EXPAND | wx.ALL, border=5)
        self.SetSizer(main_sizer)

    def load_options(self) -> None:
        checked_video_formats: list[str] = [
            VIDEO_FORMATS[get_key(vformat, VIDEO_FORMATS)]
            for vformat in self.opt_manager.options["selected_video_formats"]
        ]
        self.video_formats_checklistbox.SetCheckedStrings(checked_video_formats)
        checked_audio_formats: list[str] = [
            AUDIO_FORMATS[get_key(aformat, AUDIO_FORMATS)]
            for aformat in self.opt_manager.options["selected_audio_formats"]
        ]
        self.audio_formats_checklistbox.SetCheckedStrings(checked_audio_formats)
        self.keep_video_checkbox.SetValue(self.opt_manager.options["keep_video"])
        self.audio_quality_combobox.SetValue(
            self.AUDIO_QUALITY[self.opt_manager.options["audio_quality"]]
        )
        self.extract_audio_checkbox.SetValue(self.opt_manager.options["to_audio"])
        self.embed_thumbnail_checkbox.SetValue(
            self.opt_manager.options["embed_thumbnail"]
        )
        self.add_metadata_checkbox.SetValue(self.opt_manager.options["add_metadata"])

    def save_options(self):
        checked_video_formats: list[str] = [
            VIDEO_FORMATS[get_key(vformat, VIDEO_FORMATS)]
            for vformat in self.video_formats_checklistbox.GetCheckedStrings()
        ]
        self.opt_manager.options["selected_video_formats"] = checked_video_formats
        checked_audio_formats: list[str] = [
            AUDIO_FORMATS[get_key(aformat, AUDIO_FORMATS)]
            for aformat in self.audio_formats_checklistbox.GetCheckedStrings()
        ]
        self.opt_manager.options["selected_audio_formats"] = checked_audio_formats
        self.opt_manager.options["keep_video"] = self.keep_video_checkbox.GetValue()
        self.opt_manager.options["audio_quality"] = get_key(
            self.audio_quality_combobox.GetValue(), self.AUDIO_QUALITY, "5"
        )
        self.opt_manager.options["to_audio"] = self.extract_audio_checkbox.GetValue()
        self.opt_manager.options[
            "embed_thumbnail"
        ] = self.embed_thumbnail_checkbox.GetValue()
        self.opt_manager.options["add_metadata"] = self.add_metadata_checkbox.GetValue()


class DownloadsTab(TabPanel):

    FILESIZES: dict[str, str] = {
        "": "Bytes",
        "k": "Kilobytes",
        "m": "Megabytes",
        "g": "Gigabytes",
        "t": "Terabytes",
        "p": "Petabytes",
        "e": "Exabytes",
        "z": "Zettabytes",
        "y": "Yottabytes",
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Lang code = ISO 639-1
        self.SUBS_LANG: dict[str, str] = {
            "en": _("English"),
            "fr": _("French"),
            "de": _("German"),
            "el": _("Greek"),
            "he": _("Hebrew"),
            "it": _("Italian"),
            "pl": _("Polish"),
            "pt": _("Portuguese"),
            "ru": _("Russian"),
            "es": _("Spanish"),
            "sv": _("Swedish"),
            "tr": _("Turkish"),
            "sq": _("Albanian"),
            "zh": _("Chinese"),
        }

        self.SUBS_CHOICES: list[str] = [
            _("None"),
            _("Automatic subtitles (YOUTUBE ONLY)"),
            _("All available subtitles"),
            _("Subtitles by language"),
        ]

        self.subtitles_label = self.crt_statictext(_("Subtitles"))
        self.subtitles_combobox = self.crt_combobox(
            self.SUBS_CHOICES, event_handler=self._on_subtitles
        )
        self.subtitles_lang_listbox = self.crt_listbox(list(self.SUBS_LANG.values()))

        self.subtitles_opts_label = self.crt_statictext(_("Subtitles options"))
        self.embed_subs_checkbox = self.crt_checkbox(
            _("Embed subtitles into video file (mp4 ONLY)")
        )

        self.playlist_box = self.crt_staticbox(_("Playlist"))

        self.playlist_start_label = self.crt_statictext(_("Start"))
        self.playlist_start_spinctrl = self.crt_spinctrl((1, 9999))
        self.playlist_stop_label = self.crt_statictext(_("Stop"))
        self.playlist_stop_spinctrl = self.crt_spinctrl()
        self.playlist_max_label = self.crt_statictext(_("Max"))
        self.playlist_max_spinctrl = self.crt_spinctrl()

        self.filesize_box = self.crt_staticbox(_("Filesize"))

        self.filesize_max_label = self.crt_statictext(_("Max"))
        self.filesize_max_spinctrl = self.crt_spinctrl((0, 1024))
        self.filesize_max_sizeunit_combobox = self.crt_combobox(
            list(self.FILESIZES.values())
        )
        self.filesize_min_label = self.crt_statictext(_("Min"))
        self.filesize_min_spinctrl = self.crt_spinctrl((0, 1024))
        self.filesize_min_sizeunit_combobox = self.crt_combobox(
            list(self.FILESIZES.values())
        )

        self._set_layout()

    def _set_layout(self) -> None:
        main_sizer = wx.BoxSizer(wx.HORIZONTAL)
        vertical_sizer = wx.BoxSizer(wx.VERTICAL)

        vertical_sizer.Add(self.subtitles_label)
        vertical_sizer.Add(self.subtitles_combobox, flag=wx.EXPAND | wx.ALL, border=5)
        vertical_sizer.Add(
            self.subtitles_lang_listbox,
            1,
            wx.EXPAND | wx.LEFT | wx.RIGHT | wx.BOTTOM,
            border=5,
        )

        vertical_sizer.Add(self.subtitles_opts_label, flag=wx.TOP, border=5)
        vertical_sizer.Add(self.embed_subs_checkbox, flag=wx.ALL, border=5)

        plist_and_fsize_sizer = wx.BoxSizer(wx.HORIZONTAL)
        plist_and_fsize_sizer.Add(self._build_playlist_sizer(), 1, wx.EXPAND)
        plist_and_fsize_sizer.AddSpacer(5)
        plist_and_fsize_sizer.Add(self._build_filesize_sizer(), 1, wx.EXPAND)

        vertical_sizer.Add(plist_and_fsize_sizer, 1, wx.EXPAND | wx.TOP, border=5)

        main_sizer.Add(vertical_sizer, 1, wx.EXPAND | wx.ALL, border=5)
        self.SetSizer(main_sizer)

    def _build_playlist_sizer(self) -> wx.StaticBoxSizer:
        playlist_box_sizer = wx.StaticBoxSizer(self.playlist_box, wx.VERTICAL)
        playlist_box_sizer.AddSpacer(10)

        border = wx.GridBagSizer(5, 40)

        border.Add(self.playlist_start_label, (0, 0), flag=wx.ALIGN_CENTER_VERTICAL)
        border.Add(self.playlist_start_spinctrl, (0, 1))

        border.Add(self.playlist_stop_label, (1, 0), flag=wx.ALIGN_CENTER_VERTICAL)
        border.Add(self.playlist_stop_spinctrl, (1, 1))

        border.Add(self.playlist_max_label, (2, 0), flag=wx.ALIGN_CENTER_VERTICAL)
        border.Add(self.playlist_max_spinctrl, (2, 1))

        playlist_box_sizer.Add(border, flag=wx.ALIGN_CENTER)

        return playlist_box_sizer

    def _build_filesize_sizer(self) -> wx.StaticBoxSizer:
        filesize_box_sizer = wx.StaticBoxSizer(self.filesize_box, wx.VERTICAL)

        border = wx.GridBagSizer(5, 20)

        border.Add(self.filesize_max_label, (0, 0), (1, 2), wx.ALIGN_CENTER_HORIZONTAL)

        border.Add(self.filesize_max_spinctrl, (1, 0))
        border.Add(self.filesize_max_sizeunit_combobox, (1, 1))

        border.Add(self.filesize_min_label, (2, 0), (1, 2), wx.ALIGN_CENTER_HORIZONTAL)

        border.Add(self.filesize_min_spinctrl, (3, 0))
        border.Add(self.filesize_min_sizeunit_combobox, (3, 1))

        filesize_box_sizer.Add(border, flag=wx.ALIGN_CENTER)

        return filesize_box_sizer

    # noinspection PyUnusedLocal
    def _on_subtitles(self, event) -> None:
        """Event handler for the wx.EVT_COMBOBOX of the subtitles_combobox."""
        self.subtitles_lang_listbox.Enable(
            self.subtitles_combobox.GetValue() == self.SUBS_CHOICES[-1]
        )

    def load_options(self) -> None:
        # NOTE Find a better way to do this
        if self.opt_manager.options["write_subs"]:
            self.subtitles_combobox.SetValue(self.SUBS_CHOICES[3])
        elif self.opt_manager.options["write_all_subs"]:
            self.subtitles_combobox.SetValue(self.SUBS_CHOICES[2])
        elif self.opt_manager.options["write_auto_subs"]:
            self.subtitles_combobox.SetValue(self.SUBS_CHOICES[1])
        else:
            self.subtitles_combobox.SetValue(self.SUBS_CHOICES[0])

        self.subtitles_lang_listbox.SetStringSelection(
            self.SUBS_LANG[self.opt_manager.options["subs_lang"]]
        )
        self.embed_subs_checkbox.SetValue(self.opt_manager.options["embed_subs"])
        self.playlist_start_spinctrl.SetValue(
            self.opt_manager.options["playlist_start"]
        )
        self.playlist_stop_spinctrl.SetValue(self.opt_manager.options["playlist_end"])
        self.playlist_max_spinctrl.SetValue(self.opt_manager.options["max_downloads"])
        self.filesize_min_spinctrl.SetValue(self.opt_manager.options["min_filesize"])
        self.filesize_max_spinctrl.SetValue(self.opt_manager.options["max_filesize"])
        self.filesize_min_sizeunit_combobox.SetValue(
            self.FILESIZES[self.opt_manager.options["min_filesize_unit"]]
        )
        self.filesize_max_sizeunit_combobox.SetValue(
            self.FILESIZES[self.opt_manager.options["max_filesize_unit"]]
        )

        self._on_subtitles(None)

    def save_options(self) -> None:
        subs_choice: int = self.SUBS_CHOICES.index(self.subtitles_combobox.GetValue())

        self.opt_manager.options["write_subs"] = False
        self.opt_manager.options["write_all_subs"] = False
        self.opt_manager.options["write_auto_subs"] = False

        if subs_choice == 1:
            self.opt_manager.options["write_auto_subs"] = True
        elif subs_choice == 2:
            self.opt_manager.options["write_all_subs"] = True
        elif subs_choice == 3:
            self.opt_manager.options["write_subs"] = True

        self.opt_manager.options["subs_lang"] = get_key(
            self.subtitles_lang_listbox.GetStringSelection(), self.SUBS_LANG, "en"
        )
        self.opt_manager.options["embed_subs"] = self.embed_subs_checkbox.GetValue()
        self.opt_manager.options[
            "playlist_start"
        ] = self.playlist_start_spinctrl.GetValue()
        self.opt_manager.options[
            "playlist_end"
        ] = self.playlist_stop_spinctrl.GetValue()
        self.opt_manager.options[
            "max_downloads"
        ] = self.playlist_max_spinctrl.GetValue()
        self.opt_manager.options["min_filesize"] = self.filesize_min_spinctrl.GetValue()
        self.opt_manager.options["max_filesize"] = self.filesize_max_spinctrl.GetValue()
        self.opt_manager.options["min_filesize_unit"] = get_key(
            self.filesize_min_sizeunit_combobox.GetValue(), self.FILESIZES, ""
        )
        self.opt_manager.options["max_filesize_unit"] = get_key(
            self.filesize_max_sizeunit_combobox.GetValue(), self.FILESIZES, ""
        )


class AdvancedTab(TabPanel):

    TEXTCTRL_SIZE: tuple[int, int] = (300, -1)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.parent: OptionsFrame = args[0]
        self.retries_label = self.crt_statictext(_("Retries"))
        self.retries_spinctrl = self.crt_spinctrl((1, 999))

        self.auth_label = self.crt_statictext(_("Authentication"))

        self.username_label = self.crt_statictext(_("Username"))
        self.username_textctrl = self.crt_textctrl()
        self.password_label = self.crt_statictext(_("Password"))
        self.password_textctrl = self.crt_textctrl(wx.TE_PASSWORD)
        self.video_pass_label = self.crt_statictext(_("Video password"))
        self.video_pass_textctrl = self.crt_textctrl(wx.TE_PASSWORD)

        self.network_label = self.crt_statictext(_("Network"))

        self.proxy_label = self.crt_statictext(_("Proxy"))
        self.proxy_textctrl = self.crt_textctrl()
        self.useragent_label = self.crt_statictext(_("User agent"))
        self.useragent_textctrl = self.crt_textctrl()
        self.referer_label = self.crt_statictext(_("Referer"))
        self.referer_textctrl = self.crt_textctrl()

        self.logging_label = self.crt_statictext(_("Logging"))

        self.enable_log_checkbox = self.crt_checkbox(
            _("Enable log"), self._on_enable_log
        )
        self.view_log_button = self.crt_button(_("View"), self._on_view)
        self.clear_log_button = self.crt_button(_("Clear"), self._on_clear)

        self._set_layout()

        if self.log_manager is None:
            self.view_log_button.Disable()
            self.clear_log_button.Disable()

    def _set_layout(self):
        main_sizer = wx.BoxSizer(wx.HORIZONTAL)
        vertical_sizer = wx.BoxSizer(wx.VERTICAL)

        # Set up retries box
        retries_sizer = wx.BoxSizer(wx.HORIZONTAL)
        retries_sizer.Add(self.retries_label, flag=wx.ALIGN_CENTER_VERTICAL)
        retries_sizer.AddSpacer(20)
        retries_sizer.Add(self.retries_spinctrl)
        vertical_sizer.Add(
            retries_sizer, flag=wx.ALIGN_RIGHT | wx.TOP | wx.RIGHT, border=5
        )

        # Set up authentication box
        vertical_sizer.Add(self.auth_label, flag=wx.TOP, border=10)
        auth_sizer = wx.GridBagSizer(5, -1)

        auth_sizer.Add(self.username_label, (0, 0), flag=wx.ALIGN_CENTER_VERTICAL)
        auth_sizer.Add(self.username_textctrl, (0, 2))

        auth_sizer.Add(self.password_label, (1, 0), flag=wx.ALIGN_CENTER_VERTICAL)
        auth_sizer.Add(self.password_textctrl, (1, 2))

        auth_sizer.Add(self.video_pass_label, (2, 0), flag=wx.ALIGN_CENTER_VERTICAL)
        auth_sizer.Add(self.video_pass_textctrl, (2, 2))

        auth_sizer.AddGrowableCol(1)
        vertical_sizer.Add(auth_sizer, flag=wx.EXPAND | wx.ALL, border=5)

        # Set up network box
        vertical_sizer.Add(self.network_label, flag=wx.TOP, border=10)
        network_sizer = wx.GridBagSizer(5, -1)

        network_sizer.Add(self.proxy_label, (0, 0), flag=wx.ALIGN_CENTER_VERTICAL)
        network_sizer.Add(self.proxy_textctrl, (0, 2))

        network_sizer.Add(self.useragent_label, (1, 0), flag=wx.ALIGN_CENTER_VERTICAL)
        network_sizer.Add(self.useragent_textctrl, (1, 2))

        network_sizer.Add(self.referer_label, (2, 0), flag=wx.ALIGN_CENTER_VERTICAL)
        network_sizer.Add(self.referer_textctrl, (2, 2))

        network_sizer.AddGrowableCol(1)
        vertical_sizer.Add(network_sizer, flag=wx.EXPAND | wx.ALL, border=5)

        # Set up logging box
        vertical_sizer.Add(self.logging_label, flag=wx.TOP, border=10)

        logging_sizer = wx.BoxSizer(wx.HORIZONTAL)
        logging_sizer.Add(self.enable_log_checkbox)
        logging_sizer.AddSpacer(-1)
        logging_sizer.Add(self.view_log_button)
        logging_sizer.AddSpacer(5)
        logging_sizer.Add(self.clear_log_button)

        vertical_sizer.Add(logging_sizer, flag=wx.EXPAND | wx.ALL, border=5)

        main_sizer.Add(vertical_sizer, 1, wx.EXPAND | wx.ALL, border=5)
        self.SetSizer(main_sizer)

    # noinspection PyUnusedLocal
    def _on_enable_log(self, event) -> None:
        """Event handler for the wx.EVT_CHECKBOX of the enable_log_checkbox."""
        wx.MessageBox(
            _("In order for the changes to take effect please restart {0}").format(
                __appname__
            ),
            _("Restart"),
            wx.OK | wx.ICON_INFORMATION,
            self,
        )

    # noinspection PyUnusedLocal
    def _on_view(self, event) -> None:
        """Event handler for the wx.EVT_BUTTON of the view_log_button."""
        log_window = LogGUI(self.parent)
        log_window.load(self.log_manager.log_file)
        log_window.Show()

    # noinspection PyUnusedLocal
    def _on_clear(self, event) -> None:
        """Event handler for the wx.EVT_BUTTON of the clear_log_button."""
        if self.log_manager is not None:
            self.log_manager.clear()

    def load_options(self) -> None:
        self.retries_spinctrl.SetValue(self.opt_manager.options["retries"])
        self.username_textctrl.SetValue(self.opt_manager.options["username"])
        self.password_textctrl.SetValue(self.opt_manager.options["password"])
        self.video_pass_textctrl.SetValue(self.opt_manager.options["video_password"])
        self.proxy_textctrl.SetValue(self.opt_manager.options["proxy"])
        self.useragent_textctrl.SetValue(self.opt_manager.options["user_agent"])
        self.referer_textctrl.SetValue(self.opt_manager.options["referer"])
        self.enable_log_checkbox.SetValue(self.opt_manager.options["enable_log"])

    def save_options(self) -> None:
        self.opt_manager.options["retries"] = self.retries_spinctrl.GetValue()
        self.opt_manager.options["username"] = self.username_textctrl.GetValue()
        self.opt_manager.options["password"] = self.password_textctrl.GetValue()
        self.opt_manager.options["video_password"] = self.video_pass_textctrl.GetValue()
        self.opt_manager.options["proxy"] = self.proxy_textctrl.GetValue()
        self.opt_manager.options["user_agent"] = self.useragent_textctrl.GetValue()
        self.opt_manager.options["referer"] = self.referer_textctrl.GetValue()
        self.opt_manager.options["enable_log"] = self.enable_log_checkbox.GetValue()


class ExtraTab(TabPanel):

    CLI_BACKEND: list[str] = [YOUTUBEDL_BIN, YTDLP_BIN]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.cli_label = self.crt_statictext(_("CLI Backend"))
        self.cli_combobox = self.crt_combobox(self.CLI_BACKEND)

        self.cmdline_args_label = self.crt_statictext(
            _("CLI Backend command line options (e.g. --help)")
        )
        self.cmdline_args_textctrl = self.crt_textctrl(wx.TE_MULTILINE)

        self.extra_opts_label = self.crt_statictext(_("Extra options"))

        self.youtube_dl_debug_checkbox = self.crt_checkbox(_("Debug CLI Backend"))
        self.ignore_errors_checkbox = self.crt_checkbox(_("Ignore errors"))
        self.ignore_config_checkbox = self.crt_checkbox(_("Ignore CLI Backend config"))
        self.no_mtime_checkbox = self.crt_checkbox(_("No mtime"))
        self.native_hls_checkbox = self.crt_checkbox(_("Prefer native HLS"))

        self._set_layout()

    def _set_layout(self) -> None:
        main_sizer = wx.BoxSizer(wx.HORIZONTAL)
        vertical_sizer = wx.BoxSizer(wx.VERTICAL)

        vertical_sizer.Add(self.cli_label)
        vertical_sizer.Add(self.cli_combobox, flag=wx.EXPAND | wx.ALL, border=5)
        vertical_sizer.Add(self.cmdline_args_label, flag=wx.TOP, border=5)
        vertical_sizer.Add(self.cmdline_args_textctrl, 1, wx.EXPAND | wx.ALL, border=5)
        vertical_sizer.Add(self.extra_opts_label, flag=wx.TOP, border=5)

        extra_opts_sizer = wx.WrapSizer()
        extra_opts_sizer.Add(self.youtube_dl_debug_checkbox)
        extra_opts_sizer.AddSpacer(5)
        extra_opts_sizer.Add(self.ignore_errors_checkbox)
        extra_opts_sizer.AddSpacer(5)
        extra_opts_sizer.Add(self.ignore_config_checkbox)
        extra_opts_sizer.AddSpacer(5)
        extra_opts_sizer.Add(self.no_mtime_checkbox)
        extra_opts_sizer.AddSpacer(5)
        extra_opts_sizer.Add(self.native_hls_checkbox)

        vertical_sizer.Add(extra_opts_sizer, flag=wx.ALL, border=5)

        main_sizer.Add(vertical_sizer, 1, wx.EXPAND | wx.ALL, border=5)
        self.SetSizer(main_sizer)

    @staticmethod
    def clean_cmd_args(args: str) -> str:
        return args.replace("'", "").replace('"', "")

    def load_options(self) -> None:
        self.cli_combobox.SetValue(
            self.opt_manager.options.get("cli_backend", YOUTUBEDL_BIN)
        )
        self.cmdline_args_textctrl.SetValue(self.opt_manager.options["cmd_args"])
        self.ignore_errors_checkbox.SetValue(self.opt_manager.options["ignore_errors"])
        self.youtube_dl_debug_checkbox.SetValue(
            self.opt_manager.options["youtube_dl_debug"]
        )
        self.ignore_config_checkbox.SetValue(self.opt_manager.options["ignore_config"])
        self.native_hls_checkbox.SetValue(self.opt_manager.options["native_hls"])
        self.no_mtime_checkbox.SetValue(self.opt_manager.options["nomtime"])

    def save_options(self) -> None:
        self.opt_manager.options["cli_backend"] = self.cli_combobox.GetValue()
        self.opt_manager.options["cmd_args"] = self.cmdline_args_textctrl.GetValue()
        self.opt_manager.options[
            "ignore_errors"
        ] = self.ignore_errors_checkbox.GetValue()
        self.opt_manager.options[
            "youtube_dl_debug"
        ] = self.youtube_dl_debug_checkbox.GetValue()
        self.opt_manager.options[
            "ignore_config"
        ] = self.ignore_config_checkbox.GetValue()
        self.opt_manager.options["native_hls"] = self.native_hls_checkbox.GetValue()
        self.opt_manager.options["nomtime"] = self.no_mtime_checkbox.GetValue()
